from typing import List
import os
import sys
import subprocess
import json

from vosk import KaldiRecognizer
from pydub import AudioSegment
from tqdm import tqdm
from math import ceil

from .models import load_model
from .post_processing import post_process_text, post_process_timecoded
from ..audio import get_audiofile_length



def transcribe_segment(segment: AudioSegment) -> str:
    """ Transcribe a short AudioSegment """
    model = load_model()
    recognizer = KaldiRecognizer(model, 16000)
    recognizer.SetWords(True)
    
    data = segment.raw_data
    text = []
    i = 0
    while i + 4000 < len(data):
        if recognizer.AcceptWaveform(data[i:i+4000]):
            text.append(json.loads(recognizer.Result())["text"])
        i += 4000
    recognizer.AcceptWaveform(data[i:])
    text.append(json.loads(recognizer.FinalResult())["text"])

    return text



def transcribe_segment_timecoded(segment: AudioSegment) -> List[dict]:
    """ Transcribe a short AudioSegment, keeping the timecodes

        The resulting transcription is a list of Vosk tokens
        Each Vosk token is a dictionary of the form:
            {'word': str, 'start': float, 'end': float, 'conf': float}
        'start' and 'end' keys are in seconds
        'conf' is a normalized confidence score
    """
    model = load_model()
    recognizer = KaldiRecognizer(model, 16000)
    recognizer.SetWords(True)
    
    data = segment.get_array_of_samples().tobytes()
    timecoded_text = []
    i = 0
    while i + 4000 < len(data):
        if recognizer.AcceptWaveform(data[i:i+4000]):
            result = json.loads(recognizer.Result())
            if "result" in result:
                timecoded_text.extend(result["result"])
        i += 4000
    recognizer.AcceptWaveform(data[i:])
    result = json.loads(recognizer.FinalResult())
    if "result" in result:
        timecoded_text.extend(result["result"])
    
    return timecoded_text



def transcribe_file(filepath: str, normalize=False) -> List[str]:
    def format_output(sentence, normalize=False):
        sentence = post_process_text(sentence, normalize)
        return sentence
    
    if not os.path.exists(filepath):
        print("Couldn't find {}".format(filepath), file=sys.stderr)

    model = load_model()
    recognizer = KaldiRecognizer(model, 16000)
    recognizer.SetWords(True)
    
    text = []

    with subprocess.Popen(["ffmpeg", "-loglevel", "quiet", "-i",
                                filepath,
                                "-ar", "16000" , "-ac", "1", "-f", "s16le", "-"],
                                stdout=subprocess.PIPE) as process:

        while True:
            data = process.stdout.read(4000)
            if len(data) == 0:
                break
            if recognizer.AcceptWaveform(data):
                sentence = json.loads(recognizer.Result())["text"]
                result = format_output(sentence, normalize)
                if result:
                    text.append(result)
        sentence = json.loads(recognizer.FinalResult())["text"]
        result = format_output(sentence, normalize)
        if result:
            text.append(result)
    
    return text



def transcribe_file_timecoded(filepath: str, normalize=False) -> List[dict]:
    """ Return list of infered words with associated timecodes (vosk format)

        Parameters
            normalized (boolean): inverse-normalize sentences
        
        The resulting transcription is a list of Vosk tokens
        Each Vosk token is a dictionary of the form:
            {'word': str, 'start': float, 'end': float, 'conf': float}
        'start' and 'end' keys are in seconds
        'conf' is a normalized confidence score
    """

    def format_output(result) -> List[dict]:
        jres = json.loads(result)
        return jres.get("result", [])

    model = load_model()
    recognizer = KaldiRecognizer(model, 16000)
    recognizer.SetWords(True)
    
    total_duration = get_audiofile_length(filepath)
    progress_bar = tqdm(total=ceil(total_duration))
    i = 0
    cumul_frames = 0

    tokens = []
    with subprocess.Popen(["ffmpeg", "-loglevel", "quiet", "-i",
                                filepath,
                                "-ar", "16000" , "-ac", "1", "-f", "s16le", "-"],
                                stdout=subprocess.PIPE) as process:
        while True:
            data = process.stdout.read(4000)
            if len(data) == 0:
                break
            if recognizer.AcceptWaveform(data):
                tokens.extend(format_output(recognizer.Result()))
            cumul_frames += len(data) // 2
            if i%10 == 0:
                progress_bar.update(cumul_frames / 16000)
                cumul_frames = 0
            i += 1
        tokens.extend(format_output(recognizer.FinalResult()))
    progress_bar.close()
    
    return tokens



import os
import typing


# Proper nouns dictionary
# with phonemes when name has a foreign or particular pronunciations

proper_nouns = dict()
_proper_nouns_phon_path = __file__.replace("__init__.py", "proper_nouns_phon.tsv")

with open(_proper_nouns_phon_path, 'r') as f:
    for l in f.readlines():
        l = l.strip()
        if l.startswith('#') or not l: continue
        w, *pron = l.split(maxsplit=1)
        pron = pron if pron else []
        
        if w in proper_nouns and pron:
            proper_nouns[w].append(pron[0])
        else:
            proper_nouns[w] = pron



# Noun dictionary
# Things that you can count

nouns_f = set()
_nouns_f_path = __file__.replace("__init__.py", "noun_f.tsv")

with open(_nouns_f_path, 'r') as f:
    for l in f.readlines():
        l = l.strip()
        if l.startswith('#') or not l: continue
        nouns_f.add(l)


nouns_m = set()
_nouns_m_path = __file__.replace("__init__.py", "noun_m.tsv")

with open(_nouns_m_path, 'r') as f:
    for l in f.readlines():
        l = l.strip()
        if l.startswith('#') or not l: continue
        nouns_m.add(l)



# Acronyms dictionary

_acronyms_path = __file__.replace("__init__.py", "acronyms.tsv")

def get_acronyms_dict():
    """
        Acronyms are stored in UPPERCASE in dictionary
        Values are lists of strings for all possible pronunciation of an acronym
    """
    acronyms = dict()
    # for l in "BCDFGHIJKLMPQRSTUVWXZ":
    #     acronyms[l] = [acr2f[l]]
    
    if os.path.exists(_acronyms_path):
        with open(_acronyms_path, 'r') as f:
            for l in f.readlines():
                l = l.strip()
                if l.startswith('#') or not l: continue
                acr, pron = l.split(maxsplit=1)
                if acr in acronyms:
                    acronyms[acr].append(pron)
                else:
                    acronyms[acr] = [pron]
    else:
        print("Acronym dictionary not found... creating file")
        open(_acronyms_path, 'a').close()
    return acronyms

acronyms = get_acronyms_dict()



# Common word mistakes

corrected_tokens = dict()
_corrected_tokens_path = __file__.replace("__init__.py", "corrected_tokens.tsv")

with open(_corrected_tokens_path, 'r') as f:
    for l in f.readlines():
        l = l.strip()
        if l.startswith('#') or not l: continue
        k, v = l.split('\t')
        v = v.split()
        corrected_tokens[k] = v


# Standardization tokens

standard_tokens = dict()
_standard_tokens_path = __file__.replace("__init__.py", "standard_tokens.tsv")

with open(_standard_tokens_path, 'r') as f:
    for l in f.readlines():
        l = l.strip()
        if l.startswith('#') or not l: continue
        k, v = l.split('\t')
        v = v.split()
        standard_tokens[k] = v



import os.path
import json
import subprocess
from math import inf, ceil
from pydub import AudioSegment
# from pydub.utils import get_player_name
from pydub.silence import detect_nonsilent




def get_audiofile_info(filename):
    r = subprocess.check_output(['ffprobe', '-hide_banner', '-v', 'panic', '-show_streams', '-of', 'json', filename])
    r = json.loads(r)
    return r['streams'][0]



def get_audiofile_length(filename):
    """
        Get audio file length in seconds
    """
    info = get_audiofile_info(filename)
    if "duration" in info:
        return float(info["duration"])
    elif "tags" in info:
        # For MKV files
        h, m, s = info["tags"]["DURATION"].split(':')
        return float(h) * 3600 + float(m) * 60 + float(s)



def convert_to_wav(src: str, dst: str, verbose=True, keep_orig=True):
    """
        Convert to 16kHz mono pcm
        Validate filename

        Parameters
        ----------
            verbose (bool):
                Info text to stdout
            keep_orig (bool):
                keep  original file (or rename if src and dst are the same)
    """
    src = os.path.abspath(src)
    dst = os.path.abspath(dst)
    if src == dst:
        # Rename existing audio file
        rep, filename = os.path.split(src)
        basename, ext = os.path.splitext(filename)
        new_name = basename + "_orig" + ext
        new_src = os.path.join(rep, new_name)
        if verbose: print(f"AUDIO_CONV: renaming {filename} to {new_name}")
        os.rename(src, new_src)
        src = new_src

    if verbose:
        print(f"AUDIO_CONV: converting {src} to {dst}...")
    rep, filename = os.path.split(dst)
    dst = os.path.join(rep, filename)
    subprocess.call(['ffmpeg', '-v', 'panic',
                     '-i', src, '-acodec', 'pcm_s16le',
                     '-ac', '1', '-ar', '16000', dst])

    if not keep_orig:
        if verbose:
            print(f"AUDIO_CONV: Removing {src}")
        os.remove(src)



def convert_to_mp3(src, dst, verbose=True):
    """
        Convert to mp3
        Validate filename
    """
    if verbose:
        print(f"converting {src} to {dst}...")
    if os.path.abspath(src) == os.path.abspath(dst):
        print("ERROR: source and destination are the same, skipping")
        return -1
    rep, filename = os.path.split(dst)
    dst = os.path.join(rep, filename)
    subprocess.call(['ffmpeg', '-v', 'panic',
                     '-i', src,
                     '-ac', '1', dst])



def get_min_max_energy(segment: AudioSegment, chunk_size=100, overlap=50):
    """ Return the minimum dBFS of the segment, wich is above -inf.
        chunk_size and overlap in milliseconds
    """
    max_energy = -999
    energy_bins = dict()

    for i in range(0, len(segment), chunk_size-overlap):
        chunk = segment[i: i+chunk_size]  # Pydub will take care of overflow
        if chunk.dBFS == -inf:
            continue
        energy = round(chunk.dBFS)
        if energy in energy_bins:
            energy_bins[energy] += 1
        else:
            energy_bins[energy] = 1
        max_energy = max(max_energy, energy)
    
    bins_list = sorted(energy_bins.items())
    n_thresh = ceil(sum(energy_bins.values()) / 1000)

    for e, n in bins_list:
        if e <= -80: continue
        if n < n_thresh: continue
        min_energy = e
        break

    return min_energy, max_energy



def _recursive_split(song: AudioSegment, segment: list, max_length=15, min_silence_len=300):
    start, end = segment
    if (end-start) / 1000 < max_length:
        # Segment's length is short enough
        return [segment]

    min_e, max_e = get_min_max_energy(song[start:end])
    delta_e = max_e - min_e
    thresh = min_e + delta_e / 6.5
    sub_segments = detect_nonsilent(song[start:end], min_silence_len=min_silence_len, silence_thresh=thresh)

    # Filter out segments too short (< 0.8s)
    sub_segments = [ sub for sub in sub_segments if sub[1]-sub[0] > 800 ]

    if len(sub_segments) == 1:
        s, e = sub_segments[0]
        if e-s > max_length:
            # Sub divide a second time, if necessary
            min_e, max_e = get_min_max_energy(song[s:e])
            delta_e = max_e - min_e
            thresh = min_e + delta_e / 5
            sub2 = detect_nonsilent(song[s:e], min_silence_len=150, silence_thresh=thresh)
            return [ [start+s, start+e] for s, e in sub2 ]
        else:
            return [[start+s, start+e]]

    result = []
    for seg in sub_segments:
        result.extend(_recursive_split(song, [start + seg[0], start + seg[1]]))
    sub_segments = result

    return result



def split_to_segments(song: AudioSegment, max_length=15, min_length=5):
    """
        Return a list of sub-segments from a pydub AudioSegment
        Sub-segments are represented as 2-elements lists:
            [start, end]
        Where 'start' and 'end' are in milliseconds
    """
    segments: list = _recursive_split(song, [0, len(song)], max_length=max_length)

    # Extend segments length
    for i, seg in enumerate(segments[1:]):
        prev = segments[i]
        gap = seg[0] - prev[1]
        if 100 < gap < 1000:
            prev[1] += gap // 2 - 50
            seg[0] -= gap // 2 - 50

    # Concatenate short segments
    result = [segments[0]]
    for i, seg in enumerate(segments[1:]):
        prev = segments[i]
        prev_length = prev[1]-prev[0]
        seg_length = seg[1]-seg[0]
        gap = seg[0] - prev[1]
        
        if gap < 500 and prev_length + seg_length < min_length*1000:
            # Extend previous segment with this segment
            result[-1][1] = seg[1]
        else:
            result.append(seg)

    return result



#! /usr/bin/env python3
# -*- coding: utf-8 -*-

"""
    Programming pattern inspired by https://github.com/mideind/Tokenizer
"""


from typing import Iterator, Iterable, List, Any, Union, Set
import re

from .definitions import (
    is_noun, is_noun_f, is_noun_m, is_proper_noun,
    is_time, match_time,
    is_unit_number, match_unit_number,
    PUNCTUATION, LETTERS, SI_UNITS,
    OPENING_QUOTES, CLOSING_QUOTES,
)
from .utils import capitalize
from ..dicts import acronyms, corrected_tokens, standard_tokens



# Regular words

re_word = re.compile(r"['\-·" + LETTERS + r"]+", re.IGNORECASE)
match_word = lambda s: re_word.fullmatch(s)
is_word = lambda s: bool(match_word(s))

# Inclusive words (ex: arvester·ez)
# will always match as regular words
re_word_inclusive = re.compile(r"(['-" + LETTERS + r"]+)·([-" + LETTERS + r"]+)", re.IGNORECASE)
match_word_inclusive = lambda s: re_word_inclusive.fullmatch(s)
is_word_inclusive = lambda s: bool(match_word_inclusive(s))



# Percentage

# re_percent = re.compile(r"(\d+)%")
# match_percent = lambda s: re_percent.fullmatch(s)
# is_percent = lambda s: bool(match_percent(s))


# Ordinals

ORDINALS = {
    "1añ"  : "kentañ",
    "2vet" : "eilvet",
    "3de"  : "trede",
    "3vet" : "teirvet",
    "4e"   : "pevare",
    "4re"  : "pevare",
    "4vet" : "pedervet",
    "9vet" : "navet"
}

re_ordinal = re.compile(r"(\d+)vet")
match_ordinal = lambda s: re_ordinal.fullmatch(s)
is_ordinal = lambda s: s in ORDINALS or bool(match_ordinal(s))


ROMAN_ORDINALS = {
    "Iañ"  : "kentañ",
    "IIvet": "eilvet",
    "IIIde": "trede",
    "IIIe" : "trede",
    "IIIvet": "teirvet",
    "IVe"  : "pevare",
    "IVvet": "pedervet",
    "IXvet": "navet"
}

re_roman_ordinal = re.compile(r"([XVI]+)vet")
match_roman_ordinal = lambda s: re_roman_ordinal.fullmatch(s)
is_roman_ordinal = lambda s: s in ROMAN_ORDINALS or bool(match_roman_ordinal(s))

re_roman_number = re.compile(r"([XVI]+)")
match_roman_number = lambda s: re_roman_number.fullmatch(s)
is_roman_number = lambda s: bool(match_roman_number(s))



class Token:
    # Enum
    RAW = -1
    FIRST_IN_SENTENCE = 0
    END_SENTENCE = 1
    PUNCTUATION = 2
    WORD = 3
    NUMBER = 4
    ROMAN_NUMBER = 5
    NOUN = 6
    PROPER_NOUN = 7
    VERB = 8
    ACRONYM = 9
    ORDINAL = 10
    ROMAN_ORDINAL = 11
    TIME = 12
    UNIT = 13         # %, m, km2, kg...
    QUANTITY = 14     # a number and a unit (%, m, km2, kg, bloaz, den...)
    # PERCENT = 11
    UNKNOWN = 99

    descr = {
        RAW: "RAW",
        FIRST_IN_SENTENCE: "FIRST_IN_SENTENCE",
        END_SENTENCE: "END_SENTENCE",
        WORD: "WORD",
        NUMBER: "NUMBER",
        ROMAN_NUMBER: "ROMAN_NUMBER",
        PUNCTUATION: "PUNCTUATION",
        NOUN: "NOUN",
        PROPER_NOUN: "PROPER_NOUN",
        VERB: "VERB",
        ACRONYM: "ACRONYM",
        ORDINAL: "ORDINAL",
        ROMAN_ORDINAL: "ROMAN_ORDINAL",
        TIME: "TIME",
        UNIT: "UNIT",
        QUANTITY: "QUANTITY",
        # PERCENT: "PERCENT",
        UNKNOWN: "UNKNOWN",
    }

    def __init__(self, data: str, kind: int=RAW, *flags):
        self.data = data
        self.norm = []
        self.kind = kind
        self.flags: Set[Flag] = set(flags)
    
    def __repr__(self):
        return f"Token({repr(self.data)}, {self.descr[self.kind]})"


class Flag:
    FIRST_WORD = 1
    MASCULINE = 2
    FEMININE = 3
    INCLUSIVE = 4
    CAPITALIZED = 5
    CORRECTED = 6



def generate_raw_tokens(text_or_gen: Union[str, Iterable[str]]) -> Iterator[Token]:
    """ Generate raw tokens by splitting strings on whitespaces """
    
    if isinstance(text_or_gen, str):
        if not text_or_gen:
            return
        text_or_gen = [text_or_gen]
    
    for sentence in text_or_gen:
        for s in sentence.split():
            yield Token(s, Token.RAW)



def parse_punctuation(token_stream: Iterator[Token], **options: Any) -> Iterator[Token]:
    """ Parse a stream of raw tokens to find punctuation
    
        TODO:
            * words with a dot in the middle and more than 2 letters
                (ex: [...] fin miz Gouere.Laouen e oa [...])
            * rak,tost

    """

    # Normalize punctuation option
    norm_punct = options.pop('norm_punct', False)

    punct_stack = []

    for tok in token_stream:
        if tok.kind == Token.RAW:
            data = tok.data
            remainder = ""
            while data:
                tokens = []
                if re.search(r"\.\.+", data):
                    match = re.search(r"\.\.+", data)
                    if match.start() == 0:
                        # Ellipsis is at the beginning of the word
                        tokens.append(Token(match.group(), Token.PUNCTUATION))
                        data = data[match.end():]
                    else:
                        # Ellipsis in the middle or end of the word
                        left_part = data[:match.start()]
                        remainder = data[match.start():]
                        data = left_part
                elif data in PUNCTUATION:
                    # A single punctuation
                    tokens.append(Token(data, Token.PUNCTUATION))
                    # All data is consumed
                    data = remainder
                    remainder = ""
                elif re.match(r"([A-Z]\.)+", data):
                    # Single initial or group of initials (i.e: I.E.)
                    match = re.match(r"([A-Z]\.)+", data)
                    tokens.append(Token(data))
                    data = data[match.end():]
                    #data = remainder
                    #remainder = ""
                else:
                    # Parse left punctuation
                    while data and data[0] in PUNCTUATION:
                        tokens.append(Token(data[0], Token.PUNCTUATION))
                        data = data[1:]
                    # Parse right punctuation
                    deferred_tokens = []
                    while data and data[-1] in PUNCTUATION:
                        deferred_tokens.insert(0, Token(data[-1], Token.PUNCTUATION))
                        data = data[:-1]
                    if data:
                        # No more punctuation in word
                        tokens.append(Token(data, tok.kind))
                        # All data is consumed
                        data = remainder
                        remainder = ""
                    if deferred_tokens:
                        tokens.extend(deferred_tokens)

                
                for t in tokens:
                    if t.kind == Token.PUNCTUATION:
                        if norm_punct:
                            if t.data == '‚':   # dirty comma
                                t.norm.append(',')
                            if re.match(r"\.\.+", t.data):
                                t.norm.append('…')
                        
                        if t.data == '"':
                            if punct_stack and punct_stack[-1] == '"':
                                punct_stack.pop()
                            else:
                                # we use a single '"' char to represent every kind of quotation mark
                                # this prevents problems when mixing types of quotation marks
                                punct_stack.append('"')
                        elif t.data in OPENING_QUOTES:
                            punct_stack.append('"')
                        elif t.data in CLOSING_QUOTES:
                            if punct_stack and punct_stack[-1] == '"':
                                punct_stack.pop()
                        elif t.data == '(':
                            punct_stack.append('(')
                        elif t.data == ')':
                            if punct_stack and punct_stack[-1] == '(':
                                punct_stack.pop()

                    yield t
                    if not punct_stack and t.data in '.?!:;':
                        yield Token('', Token.END_SENTENCE)
        
        else:
            yield tok



def parse_regular_words(token_stream: Iterator[Token], **options: Any) -> Iterator[Token]:
    """ It should be called after `parse_punctuation`
    
        TODO:
            * Brezhoneg/Galleg
            * miz Gouere.Laouen e oa
    """

    # Arg options

    for tok in token_stream:
        if tok.kind == Token.RAW:
            if tok.data in acronyms:
                tok.kind = Token.ACRONYM
            elif is_word(tok.data):
                # Token is a simple and well formed word
                if is_proper_noun(tok.data):
                    tok.kind = Token.PROPER_NOUN
                else:
                    tok.kind = Token.WORD
                    if is_word_inclusive(tok.data):
                        tok.flags.add(Flag.INCLUSIVE)
            yield tok
        else:
            yield tok



def parse_numerals(token_stream: Iterator[Token]) -> Iterator[Token]:
    """ Look for various numeral forms: numbers, ordinals, units...
        It should be applied before `parse_regular_words` to accurately parse quantities

        TODO:
            * 1,20
            * €/miz
            * +40 %
            * d'ar Sul 10/10
            * ½
            * 02.98.00.00.00
            * 2003-2004
    """

    # prev_token = None
    num_concat = "" # buffer to contatenate numeral forms such as '12 000' -> '12000'
    for tok in token_stream:
        if tok.kind == Token.RAW:
            if tok.data.isdecimal():
                if not num_concat and len(tok.data) < 4:
                    num_concat += tok.data
                elif num_concat and len(tok.data) == 3:
                    num_concat += tok.data                    
                else:
                    # An full number
                    tok.kind = Token.NUMBER
            elif re.fullmatch(r"\d{1,3}(\.\d\d\d)+", tok.data):
                # A big number with dotted thousands (i.e: 12.000.000)
                tok.data = tok.data.replace('.', '')
                tok.kind = Token.NUMBER
            else:
                if is_roman_number(tok.data):
                    tok.kind = Token.ROMAN_NUMBER
                elif is_ordinal(tok.data):
                    tok.kind = Token.ORDINAL
                elif is_roman_ordinal(tok.data):
                    tok.kind = Token.ROMAN_ORDINAL
                elif is_time(tok.data):
                    # TODO: Check for token 'gm', 'g.m', 'GM'...
                    tok.kind = Token.TIME
                elif is_unit_number(tok.data):
                    # ex: "10m2"
                    number, unit = match_unit_number(tok.data).groups()
                    number = number.replace('.', '')
                    tok.kind = Token.QUANTITY
                    tok.data = f"{num_concat}{number}{unit}"
                    tok.number = num_concat + number
                    tok.unit = unit
                    if num_concat:
                        num_concat = ""
                elif tok.data in SI_UNITS:
                    if num_concat:
                        # ex: "10 s"
                        tok.kind = Token.QUANTITY
                        tok.number = num_concat
                        tok.unit = tok.data
                        tok.data = num_concat + tok.data
                        num_concat = ""
                    elif tok.data not in ('l', 'm', 't', 'g'):
                        tok.kind = Token.UNIT
                elif num_concat and is_noun(tok.data):
                    # ex: "32 bloaz"
                    tok.kind = Token.QUANTITY
                    tok.number = num_concat
                    tok.unit = tok.data
                    if is_word_inclusive(tok.data):
                        tok.flags.add(Flag.INCLUSIVE)
                    else:
                        if is_noun_f(tok.data):
                            tok.flags.add(Flag.FEMININE)
                        if is_noun_m(tok.data):
                            tok.flags.add(Flag.MASCULINE)
                    tok.data = f"{num_concat} {tok.data}"
                    num_concat = ""
                
                if num_concat:
                    yield Token(num_concat, Token.NUMBER)
                    num_concat = ""
        else:
            if num_concat:
                yield Token(num_concat, Token.NUMBER)
                num_concat = ""

        if not num_concat:
            yield tok
    
    if num_concat:
        yield(Token(num_concat, Token.NUMBER))



def correct_tokens(token_stream: Iterator[Token]) -> Iterator[Token]:
    """ Should be applied before `parse_regular_words` """

    def get_susbitution(word: str) -> List[str]:
        lowered = word.lower()
        
        if lowered in corrected_tokens:
            substitutes = corrected_tokens[lowered]
        elif lowered in standard_tokens:
            substitutes = standard_tokens[lowered]
        else:
            return []
        
        # Keep capitalization
        i = 0
        while lowered[i] not in LETTERS: i += 1
        if word[i].isupper():
            first = capitalize(substitutes[0])
            return [first] + substitutes[1:]
        else:
            return substitutes

        


    for tok in token_stream:
        if tok.kind == Token.RAW:
            lowered = tok.data.lower()
            substitutes = get_susbitution(tok.data)
            if substitutes:
                # We must keep the prepended apostrophe (there could be a substitution rule for it)
                yield from [ Token(s, Token.RAW, Flag.CORRECTED) for s in substitutes ]
            elif lowered.startswith("'") and lowered[1:] not in ('n', 'm', 'z'):
                # Remove prepended apostrophies
                # Check if there is a susbstitution rule for the remaining word
                substitutes = get_susbitution(tok.data[1:])
                if substitutes:
                    yield from [ Token(s, Token.RAW, Flag.CORRECTED) for s in substitutes ]
                else:
                    # Pass the word without the apostrophe
                    tok.data = tok.data[1:]
                    yield tok
            else:
                yield tok
        else:
            yield tok



def tokenize(text_or_gen: Union[str, Iterable[str]], **options: Any) -> Iterator[Token]:
    """
        TODO:
            * &
    """

    # Arg options
    autocorrect = options.pop('autocorrect', False)
    #standardize = options.pop('standardize', False)
    
    token_stream = generate_raw_tokens(text_or_gen)
    token_stream = parse_punctuation(token_stream, **options)
    if autocorrect:
        token_stream = correct_tokens(token_stream)
    token_stream = parse_numerals(token_stream)
    token_stream = parse_regular_words(token_stream, **options)
    # token_stream = parse_acronyms(token_stream)

    return token_stream



def detokenize(token_stream: Iterator[Token], **options: Any) -> str:
    
    # Parse options
    end_sentence = options.pop('end', '')
    # colored = options.pop("colored", False)

    parts: List[str] = []
    punct_stack = [] # Used to keep track of coupled punctuation (quotes and brackets)

    for tok in token_stream:
        data = tok.norm[0] if tok.norm else tok.data

        prefix = ''
        if tok.kind == Token.PUNCTUATION:
            if data in '!?:;':
                prefix = '\xa0' # Non-breakable space
            elif data == '"':
                if punct_stack and punct_stack[-1] == '"':
                    punct_stack.pop()
                    prefix = ''
                else:
                    punct_stack.append('"')
                    prefix = ' '
            elif data in OPENING_QUOTES:
                # we use a single '"' char to represent every kind of quotation mark
                # this prevents problems when mixing types of quotation marks
                punct_stack.append('"')
                prefix = ' '
            elif data in CLOSING_QUOTES:
                if punct_stack and punct_stack[-1] == '"':
                    punct_stack.pop()
                prefix = '\xa0' if data == '»' else ''
            elif data in '([':
                punct_stack.append(data)
                prefix = ' '
            elif data in ')':
                if punct_stack and punct_stack[-1] == '(':
                    punct_stack.pop()
                prefix = ''
            elif data in ']':
                if punct_stack and punct_stack[-1] == '[':
                    punct_stack.pop()
                prefix = ''
            elif data == '/…':
                prefix = ''
        
        elif tok.kind == Token.END_SENTENCE:
            prefix = end_sentence

        elif parts and parts[-1]:
            last_char = parts[-1][-1]
            if last_char == '«':
                prefix = '\xa0'
            elif punct_stack and last_char == punct_stack[-1]:
                prefix = ''
            elif punct_stack and last_char == '“':
                prefix = ''
            elif last_char not in '-/':
                prefix = ' '

        if parts:
            parts.append(prefix + data)
        else:
            # First word in sentence
            parts.append(data if data else '')
        if parts[-1] == '':
            parts.pop()
    
    if parts:
        ret = ''.join(parts)
        return ret
    return ''



def split_sentences(text_or_gen: Union[str, Iterable[str]], **options: Any) -> Iterator[str]:
    """ Split a line (or list of lines) according to its punctuation
        This function can be used independently

        Options:
            'end' : end the sentences with the given character 

    """

    end_sentence = options.pop("end", '\n')
    # preserve_newline = options.pop("preserve_newline", False)

    if isinstance(text_or_gen, str):
        if not text_or_gen:
            return
        text_or_gen = [text_or_gen]
    
    token_stream = generate_raw_tokens(text_or_gen)
    token_stream = parse_punctuation(token_stream)

    current_sentence = []
    for tok in token_stream:
        if tok.kind == Token.END_SENTENCE:
            yield detokenize(current_sentence, **options) + end_sentence
            current_sentence = []
        else:
            current_sentence.append(tok)
    if current_sentence:
        yield detokenize(current_sentence, **options) + end_sentence



#! /usr/bin/env python3
# -*- coding: utf-8 -*-


# import random
import re
from typing import Iterator, List, Any
from .definitions import SI_UNITS
from .tokenizer import match_time
from .tokenizer import ORDINALS, match_ordinal, is_ordinal
from .tokenizer import ROMAN_ORDINALS, match_roman_ordinal, is_roman_ordinal
from .tokenizer import tokenize, detokenize, Token
from ..dicts import nouns_f, nouns_m



substitutions = {
    "+"     : ["mui"],
    "="     : ["kevatal da"],
    "&"     : ["ha", "hag"],
    "%"     : ["dre gant"],
    "1/2"   : ["hanter", "unan war daou"],
    "3/4"   : ["tri c'hard"],
    "eurvezh/sizhun" : ["eurvezh dre sizhun"],
    "ao."   : ["aotrou"],
    "niv." : ["niverenn"],
    "g.m." : ["goude meren"],
    "GM"   : ["goude meren"],
    "St."  : ["Sant"],
    "h.a." : ["hag all"],
    "km/h" : [],
    "c'hm" : ["c'hilometr", "c'hilometrad"],
    "s.o"  : ["sellet ouzh"],
}



# MUTATIONS

def solve_mutation_number(number: int, noun: str) -> str:
    # Special rule for "bloaz"
    if noun in ("bloaz", "vloaz"):
        if number in (1, 3, 4, 5, 9, 1000, 1_000_000, 1_000_000_000):
            return "bloaz"
        elif number%10 in (1, 3, 4, 5, 9):
            return "bloaz"
        else:
            return "vloaz"
    else:
        return noun



def solve_mutation_article(article: str, noun: str) -> List[str]:
    """ ar, an, al, un, ur, ul
        Correct the article and change the noun's first letter
        according to mutation rules.

        Some homonyme nouns can be both male an female
        hence the returned list or results
        Ex: taol
    """

    results = []
    noun_first_letter = noun[0].lower()
    is_cap = noun[0].isupper()

    if noun_first_letter == 'l':
        article = article[:-1] + 'l'
    elif noun_first_letter in "adehinot":
        article = article[:-1] + 'n'
    else:
        article = article[:-1] + 'r'

    if noun == 'dor':
        # Special rule
        noun = 'n' + noun[1:]
        results.append(f"{article} {noun}")
    elif noun in nouns_m:
        if noun_first_letter == 'k':
            noun = "c'h" + noun[1:]

        if is_cap:
            noun = noun.capitalize()
        
        results.append(f"{article} {noun}")
    
    if noun in nouns_f:
        if noun_first_letter == 'k':
            noun = 'g' + noun[1:]
        elif noun_first_letter == 'p':
            noun = 'b' + noun[1:]
        elif noun_first_letter == 't':
            noun = 'd' + noun[1:]
        elif noun.lower().startswith('gw'):
            noun = 'w' + noun[2:]
        elif noun_first_letter == 'g':
            noun = "c'h" + noun[1:]
        elif noun_first_letter == 'b':
            noun = 'v' + noun[1:]
        elif noun_first_letter == 'm':
            noun = 'v' + noun[1:]
        
        if is_cap:
            noun = noun.capitalize()
        
        results.append(f"{article} {noun}")

    return results



def norm_number_noun(number: int, noun: str) -> str:
    """ (75, bloaz) -> "pemp bloaz ha tri-ugent"
        TODO:
            * diwall d'ar c'hemadurioù
    """

    noun = solve_mutation_number(number, noun)

    if number > 1000:
        thousands = (number//1000)*1000
        below_thousands = number - thousands
        if below_thousands == 0:
            return f"{num2txt(thousands)} {noun}"
        else:
            return f"{num2txt(thousands)} {norm_number_noun(below_thousands, noun)}"
    
    feminine = noun.lower() in nouns_f
    num_txt = num2txt(number, feminine)
    noun_first_letter = noun[0].lower()
    # is_cap = noun[0].isupper()
    if str(number).endswith('1'):
        if noun_first_letter == 'l':
            num_txt = num_txt.replace('unan', 'ul')
        elif noun_first_letter in "adehinot":
            num_txt = num_txt.replace('unan', 'un')
        else:
            num_txt = num_txt.replace('unan', 'ur')

    split_index = max( [num_txt.rfind(tok) for tok in (" ha ", " hag ", " warn-")] )
    if split_index == -1:
        return f"{num_txt} {noun}"
    else:
        return f"{num_txt[:split_index]} {noun}{num_txt[split_index:]}"



num_units = ["", "unan", "daou", "tri", "pevar", "pemp", "c'hwec'h", "seizh", "eizh", "nav",
             "dek", "unnek", "daouzek", "trizek", "pevarzek", "pemzek", "c'hwezek", "seitek", "triwec'h", "naontek", "ugent"]
num_units_f = ["", "unan", "div", "teir", "peder"]
num_tens = ["", "", "ugent", "tregont", "daou-ugent", "hanter-kant", "tri-ugent", "dek ha tri-ugent", "pevar-ugent", "dek ha pevar-ugent"]


def num2txt(num: int, feminine=False) -> str:
    if num == 0:
        return "mann"
        
    if feminine and num < 5:
        return num_units_f[num]

    if num <= 20:
        return num_units[num]

    if num < 100:
        t, u = divmod(num, 10)
        unit = num_units_f[u] if feminine and u in (2, 3, 4) else num_units[u]
        if u == 0:
            return num_tens[t]
        if t == 2:
            return unit + " warn-" + num_tens[t]
        if t == 5:
            return unit + " hag " + num_tens[t]
        if t in (7, 9):
            t -= 1; u += 10
            unit = num_units_f[u] if feminine and u in (2, 3, 4) else num_units[u]
        return unit + " ha " + num_tens[t]

    if num < 1000:
        h, u = divmod(num, 100)
        if u == 0: unit_str = ""
        else: unit_str = " " + num2txt(u, feminine=feminine)

        if h == 1:
            return "kant" + unit_str
        if h in (2, 3, 4, 9):
            return num_units[h] + " c'hant" + unit_str
        return num_units[h] + " kant" + unit_str

    if num < 1_000_000:
        m, u = divmod(num, 1_000)
        if u == 0: unit_str = ""
        else: unit_str = " " + num2txt(u)
        
        if m == 1:
            return "mil" + unit_str
        if m == 2:
            return num2txt(m) + " vil" + unit_str
        return num2txt(m) + " mil" + unit_str
    
    if num < 1_000_000_000:
        # Millions
        m, u = divmod(num, 1_000_000)
        if u == 0: unit_str = ""
        else: unit_str = " " + num2txt(u)

        if m == 1:
            return "ur milion" + unit_str
        if m == 2:
            return num2txt(m) + " vilion" + unit_str
        return num2txt(m) + " milion" + unit_str
    
    else:
        # Billions
        mrd, u = divmod(num, 1_000_000_000)
        if u == 0: unit_str = ""
        else: unit_str = " " + num2txt(u)

        if mrd == 1:
            return "ur miliard" + unit_str
        if mrd == 2:
            return num2txt(mrd) + " viliard" + unit_str
        return num2txt(mrd) + " miliard" + unit_str 


# Time (hours and minutes)

minutes = {
    15 : "ha kard",
    30 : "hanter",
}

def norm_time(s: str) -> List[str]:
    h, mn = map(int, match_time(s).groups(default=0))
    h_hyp = []
    if h == 0: h_hyp.extend(["hanternoz", "kreiznoz"])
    elif h == 12: h_hyp.append("kreisteiz")
    elif h > 12: h_hyp.append(num2txt(h-12, feminine=True) + " eur")
    else: h_hyp.append(num2txt(h, feminine=True) + " eur")
    
    mn_hyp = []
    if mn != 0:
        if mn in minutes: mn_hyp.append(minutes[mn])
        # if mn >= 50: mn_hyp.append(f"nemet {num2txt(60-mn)}")
        mn_hyp.append(num2txt(mn))

    results = []

    if mn > 30:
        next_hour = norm_time(str((h+1)%12) + 'e')
        if mn == 40 or mn > 50:
            for nh in next_hour:
                results.append(nh + f" nemet {num2txt(60-mn)}")
        elif mn == 45:
            for nh in next_hour:
                results.append(nh + f" nemet kard")

    for h in h_hyp:
        if mn_hyp:
            for mn in mn_hyp:
                results.append(f"{h} {mn}")
        else:
            results.append(h)
    
    # return sorted(results, key=len)
    return results


# Percentage
# norm_percent = lambda s: norm_number_noun(int(match_percent(s).group(1)), "dregant")


# Ordinals

norm_ordinal = lambda s: ORDINALS[s] if s in ORDINALS else num2txt(int(match_ordinal(s).group(1))) + "vet"


roman2br = {
    'I'    : "unan",
    "II"   : "daou",
    "III"  : "tri",
    "IIII" : "pevar",
    "IV"   : "pevar",
    "V"    : "pemp",
    "VI"   : "c'hwec'h",
    "VII"  : "seizh",
    "VIII" : "eizh",
    "IX"   : "nav",
    "X"    : "dek",
    "XI"   : "unnek",
    "XII"  : "daouzek",
    "XIII" : "trizek",
    "XIV"  : "pevarzek",
    "XV"   : "pempzek",
    "XVI"  : "c'hwezek",
    "XVII" : "seitek",
    "XVIII": "triwec'h",
    "XIX"  : "naontek",
    "IXX"  : "naontek",
    "XX"   : "ugent",
    "XXI"  : "un warn-ugent",
    "XXII" : "daou warn-ugent",
    "XXIII": "tri warn-ugent",
    "XXIV" : "pevar warn-ugent",
    "XXV"  : "pemp warn-ugent",
    "XXVI" : "c'hwec'h warn-ugent",
    "XXVII": "seizh warn-ugent",
    "XXVIII": "eizh warn-ugent",
    "XXIX" : "nav warn-ugent",
    "XXX"  : "tregont",
    "XXXI" : "un ha tregont",
    "XXXII": "daou ha tregont",
    "XXXIII": "tri ha tregont",
    "XXXIV": "pevar ha tregont",
    "XXXV" : "pemp ha tregont"
}

norm_roman_ordinal = lambda s: ROMAN_ORDINALS[s] if s in ROMAN_ORDINALS else roman2br[match_roman_ordinal(s).group(1)] + "vet"



def normalize_sentence(sentence: str, autocorrect=False) -> str:
    """
        Normalize a single sentence
        return: list of all possible normalization for the sentence
    """

    # Option: cap_first_letter ?

    propositions = []
    result = ""
    return detokenize(normalize(tokenize(sentence, autocorrect=autocorrect)))



def normalize(token_stream: Iterator[Token], **options: Any) -> Iterator[Token]:
    """ Text normalizer

        Bugs: the ordinal '3e' will be interpreted as a TIME token
    """

    # prev_tok = None
    # hold_token = False
    for tok in token_stream:
        if tok.kind == Token.PROPER_NOUN: tok.norm.append(tok.data)
        elif tok.kind == Token.WORD: tok.norm.append(tok.data.lower())
        elif tok.kind == Token.NUMBER:
            # hold_token = True
            tok.norm.append(num2txt(int(tok.data)))
        elif tok.kind == Token.ROMAN_NUMBER: tok.norm.append(roman2br[tok.data])
        elif tok.kind == Token.TIME: tok.norm.extend(norm_time(tok.data))
        elif tok.kind == Token.ORDINAL:
            tok.norm.append(norm_ordinal(tok.data))
        elif tok.kind == Token.ROMAN_ORDINAL: tok.norm.append(norm_roman_ordinal(tok.data))
        elif tok.kind == Token.QUANTITY:
            if tok.unit == '%':
                tok.norm.append(num2txt(int(tok.number)) + " dre gant")
            else:
                noun = tok.unit if tok.unit not in SI_UNITS else SI_UNITS[tok.unit][0]
                tok.norm.append(norm_number_noun(int(tok.number), noun))
        elif tok.kind == Token.UNIT:
            tok.norm.append(SI_UNITS[tok.data][0])

        # if hold_token:
        #     yield prev_tok
        # if not hold_token:
        #     yield tok
        # hold_token = False
        # prev_tok = tok
        yield tok



from typing import Iterator, Any

from .tokenizer import Token, tokenize, detokenize, split_sentences
from .normalizer import normalize, normalize_sentence
from .inverse_normalizer import inverse_normalize_sentence, inverse_normalize_timecoded
from .utils import (
    strip_punct, filter_out, capitalize, pre_process,
    extract_parenthesis_content, sentence_stats,
)
from .definitions import PUNCTUATION
from ..utils import read_file_drop_comments



def load_translation_dict(path: str) -> dict:
    translation_dict = dict()
    with open(path, 'r') as f:
        for line in f.readlines():
            line = line.strip()
            if '\t' in line:
                key, val = line.split('\t')
                translation_dict[key] = val
            else:
                translation_dict[line] = ""
    return translation_dict



def reverse_translation_dict(path: str, newpath: str) -> None:
    reversed = dict()
    for line in read_file_drop_comments(path):
        line = pre_process(line)
        if '\t' in line:
            key, val = line.split('\t')
            if ' ' in key or ' ' in val:
                print("no spaces allowed in translation dictionaries")
                continue
            if val in reversed:
                reversed[val] += ", {}".format(key)
            else:
                reversed[val] = key
    with open(newpath, 'w') as f:
        for k in sorted(reversed):
            f.write(f"{k}\t{reversed[k]}\n")



def correct_sentence(sentence: str) -> str:
    return detokenize(tokenize(sentence, autocorrect=True))



def translate(token_stream: Iterator[Token], tra_dict: dict, **options: Any) -> Iterator[Token]:
    """ Substitute tokens according to a given dictionary
        
        Keys with uppercase letters will be case-sentitive
        Keys with lowercase letters only will be case-insensitive
        Translate will operate according to the first match in the dictionary

        Key/value pairs of the translation dictionary can contain
        the '*' character to match any character
        Ex: "*a" : "*añ"    -> will change suffixes of words ending with 'a'
    """

    for tok in token_stream:
        for key, val in tra_dict.items():
            if key.startswith('*'):
                if key.endswith('*'):
                    # Match chars in the middle of words
                    expr = key.strip('*')
                    if expr in tok.data:
                        tok.data.replace(expr, val.strip('*'))
                else:
                    # Match at the end
                    expr = key.lstrip('*')
                    if tok.data.endswith(expr):
                        tok.data = tok.data[:-len(expr)] + val.lstrip('*')
            elif key.endswith('*'):
                # Match at the beginning
                expr = key.rstrip('*')
                if tok.data.startswith(expr):
                    tok.data = val.rstrip('*') + tok.data[len(expr):]
            elif key.islower():
                if key == tok.data.lower():
                    if tok.data.istitle():
                        tok.data = val.capitalize()
                    elif tok.data.isupper():
                        tok.data = val.upper()
                    else:
                        tok.data = val
                    break
            else:
                if key == tok.data:
                    tok.data = tra_dict[tok.data]
                    break
        yield tok



#!/usr/bin/env python3

import sys
import argparse
import queue
import sounddevice as sd

import static_ffmpeg
from vosk import KaldiRecognizer

from anaouder.asr.models import load_model, DEFAULT_MODEL
from anaouder.asr.post_processing import post_process_text
from anaouder.text import tokenize, detokenize, load_translation_dict, translate
from anaouder.version import VERSION



def int_or_str(text):
	"""Helper function for argument parsing."""
	try:
		return int(text)
	except ValueError:
		return text


def callback(indata, frames, time, status):
	"""This is called (from a separate thread) for each audio block."""
	if status:
		print(status, file=sys.stderr)
	q.put(bytes(indata))


def format_output(sentence, normalize=False, keep_fillers=False):
	sentence = post_process_text(sentence, normalize, keep_fillers)
	for td in translation_dicts:
		sentence = detokenize( translate(tokenize(sentence), td) )
	return sentence


def main_mikro() -> None:
	""" mikro cli entry point """

	global q
	global translation_dicts

	parser = argparse.ArgumentParser(add_help=False)
	parser.add_argument('-l', '--list-devices', action='store_true',
		help='show list of audio devices and exit')
	args, remaining = parser.parse_known_args()
	if args.list_devices:
		print(sd.query_devices())
		parser.exit(0)
	parser = argparse.ArgumentParser(
		description=__doc__,
		formatter_class=argparse.RawDescriptionHelpFormatter,
		parents=[parser])
	parser.add_argument("-o", "--output", type=str, metavar='FILENAME',
		help='text file to store transcriptions')
	parser.add_argument('-m', '--model', type=str, metavar='MODEL_PATH', default=DEFAULT_MODEL,
		help='Path to the model')
	parser.add_argument('-d', '--device', type=int_or_str,
		help='input device (numeric ID or substring)')
	parser.add_argument('-r', '--samplerate', type=int,
		help='sampling rate')
	parser.add_argument('-n', '--normalize', action="store_true",
		help="Normalize numbers")
	parser.add_argument("--translate", nargs='+',
		help="Use additional translation dictionaries")
	parser.add_argument("--keep-fillers", action="store_true",
		help="Keep verbal fillers ('euh', 'beñ', 'alors', 'kwa'...)")
	parser.add_argument("-v", "--version", action="version", version=f"%(prog)s v{VERSION}")
	args = parser.parse_args(remaining)

	# Use static_ffmpeg instead of ffmpeg
	static_ffmpeg.add_paths()
	
	q = queue.Queue()

	try:
		if args.samplerate is None:
			device_info = sd.query_devices(args.device, 'input')
			# soundfile expects an int, sounddevice provides a float:
			args.samplerate = int(device_info['default_samplerate'])

		model = load_model(args.model)

		translation_dicts = []
		if args.translate:
			translation_dicts = [ load_translation_dict(path) for path in args.translate ]

		if args.output:
			dump_fn = open(args.output, "w")
		else:
			dump_fn = None

		with sd.RawInputStream(samplerate=args.samplerate, blocksize = 1024, device=args.device, dtype='int16',
								channels=1, latency='high', callback=callback):
				print('#' * 80)
				print('Press Ctrl+C to stop the recording')
				print('#' * 80)
				
				rec = KaldiRecognizer(model, args.samplerate)
				
				while True:
					data = q.get()
					if rec.AcceptWaveform(data):
						result = eval(rec.Result())["text"]
						if len(result) > 0:
							print(format_output(result, normalize=args.normalize, keep_fillers=args.keep_fillers))
							if dump_fn:
								dump_fn.write(format_output(result, normalize=args.normalize, keep_fillers=args.keep_fillers)+'\n')

	except KeyboardInterrupt:
		print('\nDone')
		parser.exit(0)
	except Exception as e:
		parser.exit(type(e).__name__ + ': ' + str(e))


if __name__ == "__main__":
	main_mikro()



import os
from typing import List, Optional
from ..text.inverse_normalizer import inverse_normalize_sentence, inverse_normalize_timecoded
from ..text.definitions import is_noun
from ..utils import read_file_drop_comments



# Verbal fillers with phonetization

verbal_fillers = {
    'euh'   :   'OE',
    'euhm'  :   'OE M',
    'beñ'   :   'B EN',
    'beh'   :   'B E',
    'eba'   :   'E B A',
    'ebeñ'  :   'E B EN',
    'kwa'   :   'K W A',
    'hañ'   :   'H AN',
    'heñ'   :   'EN',
    'boñ'   :   'B ON',
    'bah'   :   'B A',
    'feñ'   :   'F EN',
    'enfin' :   'AN F EN',
    'tiens' :   'T I EN',
    'alors' :   'A L OH R',
    'allez' :   'A L E',
    'voilà' :   'V O A L A',
    'pff'   :   'P F F',
    'mais'  :   'M EH',
    'hmm'   :   'M M',
    #'oh'    :   'O',
    #'ah'    :   'A',
}



def load_postproc_dict(filepath: str) -> List[dict]:
    """ Load the post processing dictionary from a tsv file
        Keys are treated uncased
    """

    trigrams = dict()
    bigrams = dict()
    monograms = dict()
    for l in read_file_drop_comments(filepath):
        k, v = l.split('\t')
        k = tuple([t.lower() for t in k.split()])
        v = tuple(v.split())
        if len(k) == 1:
            monograms[k] = v
        elif len(k) == 2:
            bigrams[k] = v
        elif len(k) == 3:
            trigrams[k] = v
        else:
            print("ERROR (postproc_dict): number of tokens should be between 1 and 3")
            print(k)
    
    return [monograms, bigrams, trigrams]


_postproc_dict_path = os.path.join(os.path.split(__file__)[0], "postproc_sub.tsv")
_postproc_dict = load_postproc_dict(_postproc_dict_path)

_inorm_units_dict_path = os.path.join(os.path.split(__file__)[0], "inorm_units.tsv")
_inorm_units_dict = load_postproc_dict(_inorm_units_dict_path)



def post_process_text(sentence: str, normalize=False, keep_fillers=True) -> str:
    """ Apply post-processing on raw text """

    # Verbal fillers removal
    if not keep_fillers:
        sentence = sentence.split()
        parsed = []
        for word in sentence:
            if not word.lower() in verbal_fillers:
                parsed.append(word)
        sentence = ' '.join(parsed)

    sentence = apply_post_process_dict_text(sentence, _postproc_dict)
    
    # Add hyphens for "-se" and "-mañ"
    sentence = sentence.split()
    parsed = []
    prev_word = ''
    for word in sentence:
        if word in ("se", "mañ") and is_noun(prev_word):
            parsed.append('-'.join([parsed.pop(), word]))
            prev_word = parsed[-1]
        else:
            parsed.append(word)
            prev_word = word
    sentence = ' '.join(parsed)

    if normalize:
        sentence = apply_post_process_dict_text(sentence, _inorm_units_dict)
        sentence = inverse_normalize_sentence(sentence)
    return sentence



def post_process_timecoded(
        tokens: List[dict],
        normalize=False,
        keep_fillers=True) -> List[dict]:
    """ Apply post-processing on Vosk formatted result (keeping timecodes)

        Add hypens (-se, -mañ)
        Common words substitution  (optional)
        Inverse-normalization      (optional)
    """
    
    # Verbal fillers removal
    if not keep_fillers:
        parsed = []
        for idx, tok in enumerate(tokens):
            if not tok["word"].lower() in verbal_fillers:
                parsed.append(tok)
        tokens = parsed

    tokens = apply_post_process_dict_timecoded(tokens, _postproc_dict)

    # Add hyphens for "-se" and "-mañ"
    parsed = []
    prev_word = ''
    for idx, tok in enumerate(tokens):
        if tok["word"] in ("se", "mañ") and is_noun(prev_word):
            word = prev_word + '-' + tok["word"]
            new_token = {
                        "word": word,
                        "start": tokens[idx-1]["start"],
                        "end": tok["end"],
                        "conf": tok["conf"]
                        }
            parsed.append(new_token)
            prev_word = word
        else:
            parsed.append(tok)
    tokens = parsed

    if normalize:
        tokens = apply_post_process_dict_timecoded(tokens, _inorm_units_dict)
        tokens = inverse_normalize_timecoded(tokens)
    return tokens



def apply_post_process_dict_text(sentence: str, ngram_dicts: List[dict]=_postproc_dict) -> str:

    def check_ngram(n: int):
        ngram_lowered = tuple( [ t.lower() for t in tokens[idx:idx+n] ] )
        sub = ngram_dicts[n-1].get(ngram_lowered, None)
        if sub:
            for t in sub:
                translated.append(t)    #XXX the translation is always lowercase
            return True
        return False
    
    tokens = sentence.split()
    translated = []
    tlen = len(tokens)
    idx = 0
    while idx < tlen:
        if idx <= tlen-3 and check_ngram(3):
            idx += 3
            continue
        if idx <= tlen-2 and check_ngram(2):
            idx += 2
            continue
        if idx <= tlen-1 and check_ngram(1):
            pass
        else:
            translated.append(tokens[idx])
        idx += 1

    return ' '.join(translated)



def apply_post_process_dict_timecoded(tokens: List[dict], ngram_dicts: List[dict]=_postproc_dict) -> List[dict]:

    def check_ngram(n: int):
        ngram = tuple( [ t["word"].lower() for t in tokens[idx:idx+n] ] )
        sub = ngram_dicts[n-1].get(ngram, None)
        if sub:
            t_start = tokens[idx]["start"]
            t_dur = tokens[idx+n-1]["end"] - t_start
            t_dur = t_dur / len(sub)
            conf = tokens[idx]["conf"]
            for t in sub:
                token = {"word": t, "start": t_start, "end": t_start + t_dur, "conf": conf}
                translated.append(token)
                t_start += t_dur
            return True
        return False
    
    translated = []
    tlen = len(tokens)
    idx = 0
    while idx < tlen:
        if idx <= tlen-3 and check_ngram(3):
            idx += 3
            continue
        if idx <= tlen-2 and check_ngram(2):
            idx += 2
            continue
        if idx <= tlen-1 and check_ngram(1):
            pass
        else:
            translated.append(tokens[idx])
        idx += 1

    return translated


from typing import List
import os

# For eaf (Elan) file conversion
import datetime, pytz
from xml.dom import minidom

from anaouder.audio import convert_to_mp3
from anaouder.version import VERSION



def list_files_with_extension(ext, rep, recursive=True) -> List[str]:
    file_list = []
    if os.path.isdir(rep):
        for filename in os.listdir(rep):
            filename = os.path.join(rep, filename)
            if os.path.isdir(filename) and recursive:
                file_list.extend(list_files_with_extension(ext, filename))
            elif os.path.splitext(filename)[1] == ext:
                file_list.append(filename)
    return file_list



def read_file_drop_comments(path: str) -> List[str]:
    lines = []
    with open(path, 'r') as f:
        for l in f.readlines():
            l = l.strip()
            if l and not l.startswith('#'):
                lines.append(l)
    return lines



def write_eaf(segments, sentences, audiofile, type="wav"):
    """ Export to eaf (Elan) file """

    record_id = os.path.splitext(os.path.abspath(audiofile))[0]
    if type == "mp3":
        mp3_file = os.path.extsep.join((record_id, 'mp3'))
        if not os.path.exists(mp3_file):
            convert_to_mp3(audiofile, mp3_file)
        audiofile = mp3_file

    doc = minidom.Document()

    root = doc.createElement('ANNOTATION_DOCUMENT')
    root.setAttribute('AUTHOR', f'Anaouder {VERSION} (Gweltaz DG)')
    root.setAttribute('DATE', datetime.datetime.now(pytz.timezone('Europe/Paris')).isoformat(timespec='seconds'))
    root.setAttribute('FORMAT', '3.0')
    root.setAttribute('VERSION', '3.0')
    root.setAttribute('xmlns:xsi', 'http://www.w3.org/2001/XMLSchema-instance')
    root.setAttribute('xsi:noNamespaceSchemaLocation', 'http://www.mpi.nl/tools/elan/EAFv3.0.xsd')
    doc.appendChild(root)

    header = doc.createElement('HEADER')
    header.setAttribute('MEDIA_FILE', '')
    header.setAttribute('TIME_UNITS', 'milliseconds')
    root.appendChild(header)

    media_descriptor = doc.createElement('MEDIA_DESCRIPTOR')
    media_descriptor.setAttribute('MEDIA_URL', 'file://' + os.path.abspath(audiofile))
    if type == "mp3":
        media_descriptor.setAttribute('MIME_TYPE', 'audio/mpeg')
    else:
        media_descriptor.setAttribute('MIME_TYPE', 'audio/x-wav')
    media_descriptor.setAttribute('RELATIVE_MEDIA_URL', './' + os.path.basename(audiofile))
    header.appendChild(media_descriptor)

    time_order = doc.createElement('TIME_ORDER')
    last_t = 0
    for i, (s, e) in enumerate(segments):
        if s < last_t:
            s = last_t
        last_t = s
        time_slot = doc.createElement('TIME_SLOT')
        time_slot.setAttribute('TIME_SLOT_ID', f'ts{2*i+1}')
        time_slot.setAttribute('TIME_VALUE', str(s))
        time_order.appendChild(time_slot)
        time_slot = doc.createElement('TIME_SLOT')
        time_slot.setAttribute('TIME_SLOT_ID', f'ts{2*i+2}')
        time_slot.setAttribute('TIME_VALUE', str(e))
        time_order.appendChild(time_slot)
    root.appendChild(time_order)

    tier_trans = doc.createElement('TIER')
    tier_trans.setAttribute('LINGUISTIC_TYPE_REF', 'transcript')
    tier_trans.setAttribute('TIER_ID', 'Transcription')

    for i, sentence in enumerate(sentences):
        annotation = doc.createElement('ANNOTATION')
        alignable_annotation = doc.createElement('ALIGNABLE_ANNOTATION')
        alignable_annotation.setAttribute('ANNOTATION_ID', f'a{i+1}')
        alignable_annotation.setAttribute('TIME_SLOT_REF1', f'ts{2*i+1}')
        alignable_annotation.setAttribute('TIME_SLOT_REF2', f'ts{2*i+2}')
        annotation_value = doc.createElement('ANNOTATION_VALUE')
        #text = doc.createTextNode(get_cleaned_sentence(sentence, rm_bl=True, keep_dash=True, keep_punct=True)[0])
        text = doc.createTextNode(sentence.replace('*', ''))
        annotation_value.appendChild(text)
        alignable_annotation.appendChild(annotation_value)
        annotation.appendChild(alignable_annotation)
        tier_trans.appendChild(annotation)
    root.appendChild(tier_trans)

    linguistic_type = doc.createElement('LINGUISTIC_TYPE')
    linguistic_type.setAttribute('GRAPHIC_REFERENCES', 'false')
    linguistic_type.setAttribute('LINGUISTIC_TYPE_ID', 'transcript')
    linguistic_type.setAttribute('TIME_ALIGNABLE', 'true')
    root.appendChild(linguistic_type)

    language = doc.createElement('LANGUAGE')
    language.setAttribute("LANG_ID", "bre")
    language.setAttribute("LANG_LABEL", "Breton (bre)")
    root.appendChild(language)

    constraint_list = [
        ("Time_Subdivision", "Time subdivision of parent annotation's time interval, no time gaps allowed within this interval"),
        ("Symbolic_Subdivision", "Symbolic subdivision of a parent annotation. Annotations refering to the same parent are ordered"),
        ("Symbolic_Association", "1-1 association with a parent annotation"),
        ("Included_In", "Time alignable annotations within the parent annotation's time interval, gaps are allowed")
    ]
    for stereotype, description in constraint_list:
        constraint = doc.createElement('CONSTRAINT')
        constraint.setAttribute('DESCRIPTION', description)
        constraint.setAttribute('STEREOTYPE', stereotype)
        root.appendChild(constraint)

    xml_str = doc.toprettyxml(indent ="\t", encoding="UTF-8")

    return xml_str.decode("utf-8")


import os
from typing import List
from .definitions import is_noun



token_value = {
    "mann" : 0, "zero" : 0,
    "un" : 1, "ur" : 1, "ul" : 1, "unan" : 1,
    "daou" : 2, "zaou" : 2, "div" : 2,
    "tri" : 3, "teir" : 3,
    "pevar" : 4, "peder" : 4,
    "pemp" : 5, "bemp" : 5,
    "c'hwec'h" : 6,
    "seizh" : 7,
    "eizh" : 8,
    "nav" : 9,
    "dek" : 10,
    "unnek" : 11,
    "daouzek" : 12,
    "trizek" : 13,
    "pevarzek" : 14,
    "pemzek" : 15,
    "c'hwezek" : 16,
    "seitek" : 17,
    "triwec'h" : 18,
    "naontek" : 19,
    "ugent" : 20,
    "tregont" : 30,
    "kant" : 100, "c'hant" : 100,
    "mil" : 1000, "vil" : 1000,
    "milion" : 1_000_000, "vilion" : 1_000_000,
    "miliard" : 1_000_000_000, "viliard" : 1_000_000_000,
    "hanter" : 0.5,
    "ha" : '+', "hag" : '+', "warn" : '+'
}


# A star character represents any noun
numtok_chain = {
	'[': ['zero', 'unan', 'un', 'ul', 'ur', 'daou', 'zaou', 'div', 'tri', 'teir', 'pevar', 'peder', 'pemp', "bemp", "c'hwec'h", 'seizh', 'eizh', 'nav', 'dek', 'unnek', 'daouzek', 'trizek', 'pevarzek', 'pemzek', "c'hwezek", 'seitek', "triwec'h", 'naontek', 'ugent', 'tregont', 'kant', 'mil', 'hanter'],
	'zero': [']'],
	'unan': [']', 'warn', 'ha', 'hag'],
	'ur': ["c'hant", 'mil', 'milion', 'miliard', '*'],
    'un': [']', 'warn', 'ha', 'hag', '*'],
    'ul': [']', 'warn', 'ha', 'hag', '*'],
	'daou': [']', 'warn', 'ha', 'ugent', 'hag', "c'hant", 'vil', 'vilion', 'viliard', '*'],
    'zaou': [']', 'warn', 'ha', 'ugent', 'hag', "c'hant", 'vil', 'vilion', 'viliard', '*'],
    'div': [']', 'warn', 'ha', 'hag', '*'],
	'tri': [']', 'warn', 'ha', 'hag', 'ugent', "c'hant", 'mil', 'milion', '*'],
    'teir': [']', 'warn', 'ha', 'hag', '*'],
	'pevar': [']', 'warn', 'ha', 'hag', 'ugent', "c'hant", 'mil', 'milion', '*'],
    'peder': [']', 'warn', 'ha', 'hag', '*'],
	'pemp': [']', 'warn', 'ha', 'hag', 'kant', 'mil', 'milion', '*'],
    'bemp': [']', 'warn', 'ha', 'hag', 'kant', 'mil', 'milion', '*'],
	"c'hwec'h": [']', 'warn', 'ha', 'hag', 'kant', 'mil', 'milion', '*'],
	'seizh': [']', 'warn', 'ha', 'hag', 'kant', 'mil', 'milion', '*'],
	'eizh': [']', 'warn', 'ha', 'hag', 'kant', 'mil', 'milion', '*'],
	'nav': [']', 'warn', 'ha', 'hag', "c'hant", 'mil', 'milion', '*'],
	'dek': [']', 'ha', 'kant', 'mil', 'milion', '*'],
	'unnek': [']', 'ha', 'kant', 'mil', 'milion', '*'],
	'daouzek': [']', 'ha', 'kant', 'mil', 'milion', '*'],
	'trizek': [']', 'ha', 'kant', 'mil', 'milion', '*'],
	'pevarzek': [']', 'ha', 'kant', 'mil', 'milion', '*'],
	'pemzek': [']', 'ha', 'kant', 'mil', 'milion', '*'],
	"c'hwezek": [']', 'ha', 'kant', 'mil', 'milion', '*'],
	'seitek': [']', 'ha', 'kant', 'mil', 'milion', '*'],
	"triwec'h": [']', 'ha', 'kant', 'mil', 'milion', '*'],
	'naontek': [']', 'ha', 'kant', 'mil', 'milion', '*'],
	'ugent': [']', 'mil', 'milion', 'mil', 'milion', '*'],
	'tregont': [']', 'mil', 'milion', 'mil', 'milion', '*'],
	'kant': [']', 'unan', 'un', 'ul', 'ur', 'daou', 'tri', 'pevar', 'pemp', "c'hwec'h", 'seizh', 'eizh', 'nav', 'dek', 'unnek', 'daouzek', 'trizek', 'pevarzek', 'pemzek', "c'hwezek", 'seitek', "triwec'h", 'naontek', 'ugent', 'tregont', 'hanter', 'mil', 'milion', '*'],
	"c'hant": [']', 'unan', 'un', 'ul', 'ur', 'daou', 'tri', 'pevar', 'pemp', "c'hwec'h", 'seizh', 'eizh', 'nav', 'dek', 'unnek', 'daouzek', 'trizek', 'pevarzek', 'pemzek', "c'hwezek", 'seitek', "triwec'h", 'naontek', 'ugent', 'tregont', 'hanter', 'mil', 'milion', '*'],
	'mil': [']', 'unan', 'un', 'ul', 'ur', 'daou', 'tri', 'pevar', 'pemp', "c'hwec'h", 'seizh', 'eizh', 'nav', 'dek', 'unnek', 'daouzek', 'trizek', 'pevarzek', 'pemzek', "c'hwezek", 'seitek', "triwec'h", 'naontek', 'ugent', 'tregont', 'hanter', 'kant', '*'],
	'vil': [']', 'unan', 'un', 'ul', 'ur', 'daou', 'tri', 'pevar', 'pemp', "c'hwec'h", 'seizh', 'eizh', 'nav', 'dek', 'unnek', 'daouzek', 'trizek', 'pevarzek', 'pemzek', "c'hwezek", 'seitek', "triwec'h", 'naontek', 'ugent', 'tregont', 'hanter', 'kant', '*'],
	'milion': [']', '*'], 'vilion': [']', '*'],
	'miliard': [']', '*'], 'viliard': [']', '*'],
	'ha': ['tregont', 'daou', 'tri', 'pevar'],
	'hag': ['hanter'],
	'hanter': ['kant'],
	'warn': ['ugent'],
}



def solve_num_tokens(numerical_tokens: List[float]) -> float:
    # Token 0.5 ("hanter") takes precedence and is applied to next closest token
    while 0.5 in numerical_tokens:
        i = numerical_tokens.index(0.5)
        if i < len(numerical_tokens) - 1:
            numerical_tokens = numerical_tokens[:i] + [0.5*numerical_tokens[i+1]] + numerical_tokens[i+2:]
        else:
            break

    # Find highest value token
    i_max, val_max = -1, -1
    i_token_add, token_add = -1, False
    for i, val in enumerate(numerical_tokens):
        if val == '+':
            token_add = True
            i_token_add = i
        elif val > val_max:
            val_max = val
            i_max = i
    
    if token_add and val_max < 100:
        # Invert two parts of token_list around '+' symbol and solve
        inverted = numerical_tokens[i_token_add+1:] + numerical_tokens[:i_token_add]
        return solve_num_tokens(inverted)
    else:
        # solve recursively
        if len(numerical_tokens) == 0:
            return 0
        elif len(numerical_tokens) == 1:
            return numerical_tokens[0]
        else:
            left_part = solve_num_tokens(numerical_tokens[:i_max])
            if left_part == 0:
                left_part = 1
            right_part = solve_num_tokens(numerical_tokens[i_max+1:])
            return left_part * val_max + right_part



def inverse_normalize_sentence(sentence: str, min_num=5) -> str:
    """ Translate spelled numbers to more readable numbers
        This is a simple function expected on plain sentences with no punctuation
        The sentence is simply split on whitespaces

        Parameter
        ---------
            min_num (int):
                numbers below `min_num` won't be normalized
    """

    def flush_number():
        nonlocal num_tokens
        nonlocal noun
        num = solve_num_tokens( [token_value[t] for t in num_tokens] )
        if num >= min_num:
            translated.append(str(int(num)))
        else:
            translated.extend(num_tokens)
        if noun:
            translated.append(noun)
            noun = None
        num_tokens = []

    
    starters = numtok_chain['[']
    sentence = sentence.replace("-ugent", " ugent")
    sentence = sentence.replace("-kant", " kant")
    tokens = sentence.split()
    num_tokens = []
    in_chain = False
    noun = None
    translated = []
    for idx, tok in enumerate(tokens):
        if not in_chain and tok in starters:
            # Beginning of a numerical chain of words
            in_chain = True
            num_tokens = [tok]
        
        elif in_chain:
            prev = num_tokens[-1]
            if tok in numtok_chain[prev]:
                # Follow the chain of words describing a number
                num_tokens.append(tok)
            
            elif tok in starters:
                # Check for ill-formed numerical sequences,
                if ']' not in numtok_chain[prev]:
                    # treat as 2 numerical sequence in that case
                    # rewind and flush the first part
                    i = len(num_tokens)-1
                    while i >= 0 and num_tokens[i] not in ("ha", "hag"):
                        i -= 1
                    next_num_tokens = num_tokens[i+1:]
                    conj = num_tokens[i]
                    num_tokens = num_tokens[:i]
                    flush_number()
                    translated.append(conj)
                    num_tokens = next_num_tokens + [tok]
                else:
                    flush_number()
                    num_tokens = [tok]
            
            elif  '*' in numtok_chain[prev] and is_noun(tok) or tok in ('%',):
                if noun:
                    # There could be an ambiguity if another noun has already been found
                    # ex: "tri kazh ha daou besk"
                    # In that case...
                    i = len(num_tokens)-1
                    while i >= 0 and num_tokens[i] not in ("ha", "hag"):
                        i -= 1
                    next_num_tokens = num_tokens[i+1:]
                    conj = num_tokens[i]
                    num_tokens = num_tokens[:i]
                    flush_number()
                    translated.append(conj)
                    num_tokens = next_num_tokens
                    noun = tok
                elif len(tokens) > idx+1 and tokens[idx+1] in ('ha', 'hag', 'warn'):
                    noun = tok
                else:
                    # final noun
                    noun = tok
                    flush_number()
                    in_chain = False
            
            else:
                flush_number()
                translated.append(tok)
                in_chain = False

        else:
            translated.append(tok)
    
    if in_chain:
        flush_number()

    return ' '.join(translated)



def inverse_normalize_timecoded(tokens: List[dict], min_num=5) -> List[dict]:
    """ Translate spelled numbers to more readable numbers
        Same functionality as the function `inverse_normalise_sentence` but
        works with vosk tokens (embedding timecodes and confidence values)

        Parameter
        ---------
            min_num (int):
                numbers below `min_num` won't be normalized
    """

    def flush_number():
        nonlocal num_tokens
        nonlocal noun
        num = solve_num_tokens( [token_value[t["word"]] for t in num_tokens] )
        if num >= min_num:
            # Create a new token for this number
            t_start = num_tokens[0]["start"]
            t_end = num_tokens[-1]["end"]
            word = str(int(num))
            vosk_token = {"word": word, "start": t_start, "end": t_end, "conf": 1.0}
            translated.append(vosk_token)
        else:
            translated.extend(num_tokens)
        if noun:
            if noun["start"] < translated[-1]["end"]:
                # Order of tokens has been changed, correct timecodes
                dur = noun["end"] - noun["start"]
                translated[-1]["end"] -= dur
                noun["start"] = translated[-1]["end"]
                noun["end"] = noun["start"] + dur
            translated.append(noun)
            noun = None
        num_tokens = []
    

    # Start by splitting numerical words containing hyphens
    parsed_tokens = []
    for tok in tokens:
        word = tok["word"]
        if word.endswith("-ugent") or word == "hanter-kant":
            w1, w2 = word.split('-')
            t_start = tok["start"]
            t_end = tok["end"]
            dur = (t_end - t_start) / 2
            t1 = {"word": w1, "start": t_start, "end": t_start+dur, "conf": tok["conf"]}
            t2 = {"word": w2, "start": t_start+dur, "end": t_end, "conf": tok["conf"]}
            parsed_tokens.append(t1)
            parsed_tokens.append(t2)
        else:
            parsed_tokens.append(tok)
    tokens = parsed_tokens


    starters = numtok_chain['[']
    num_tokens = []
    translated = []
    in_chain = False
    noun = None
    for idx, tok in enumerate(tokens):
        if not in_chain and tok["word"] in starters:
            # Beginning of a chain of words describing a number
            in_chain = True
            num_tokens = [tok]

        elif in_chain:
            prev = num_tokens[-1]["word"]
            if tok["word"] in numtok_chain[prev]:
                # Follow the chain of words describing a number
                num_tokens.append(tok)
            
            elif tok["word"] in starters:
                # Could be the beginning of a new chain
                # Check for ill-formed numerical sequences
                if ']' not in numtok_chain[prev]:
                    # treat as 2 numerical sequence in that case
                    # rewind and flush the first part
                    i = len(num_tokens)-1
                    while i >= 0 and num_tokens[i]["word"] not in ("ha", "hag"):
                        i -= 1
                    next_num_tokens = num_tokens[i+1:]
                    conj = num_tokens[i]
                    num_tokens = num_tokens[:i]
                    flush_number()
                    translated.append(conj)
                    num_tokens = next_num_tokens + [tok]
                else:
                    flush_number()
                    num_tokens = [tok]
            
            elif  '*' in numtok_chain[prev] and is_noun(tok["word"]) or tok["word"] in ('%',):
                if noun:
                    # There could be an ambiguity if another noun has already been found
                    # ex: "tri kazh ha daou besk"
                    # In that case, flush the first part
                    i = len(num_tokens)-1
                    while i >= 0 and num_tokens[i]["word"] not in ("ha", "hag"):
                        i -= 1
                    next_num_tokens = num_tokens[i+1:]
                    conj = num_tokens[i]
                    num_tokens = num_tokens[:i]
                    flush_number()
                    translated.append(conj)
                    num_tokens = next_num_tokens
                    noun = tok
                elif len(tokens) > idx+1 and tokens[idx+1]["word"] in ('ha', 'hag', 'warn'):
                    noun = tok
                else:
                    # final noun
                    noun = tok
                    flush_number()
                    in_chain = False
            
            else:
                flush_number()
                translated.append(tok)
                in_chain = False

        else:
            translated.append(tok)
    
    if in_chain:
        flush_number()

    return translated


from typing import List, Iterator, Any, Tuple
import re
from .definitions import LETTERS, PUNCTUATION




def strip_punct(word: str) -> str:
    """ strip punctuation left and right of a word """

    while word and word[0] in PUNCTUATION:
        word = word[1:]
    while word and word[-1] in PUNCTUATION:
        word = word[:-1]
    return word



def filter_out(text: str, chars: str) -> str:
    """ Remove characters from a string """

    filtered_text = ""
    for l in text:
        if not l in chars: filtered_text += l
    return filtered_text



def capitalize(word: str) -> str:
    """ Capitalize a single word or the first word of a given sentence """
    # Should we capitalize the whole "C'H" character ? For now we don't.

    # Find first letter
    lowered = word.lower()
    i = 0
    while i < len(lowered) and lowered[i] not in LETTERS: i += 1
    if i >= len(word):
        return word
    
    capitalized = word[:i] + word[i].upper() + word[i+1:]
    return capitalized



_PATTERN_BETWEEN_PARENTHESIS = re.compile(r"\((.+?)\)")

def extract_parenthesis_content(txt: str) -> Tuple[str, str]:
    extracted = []
    remaining = txt
    match = re.search(_PATTERN_BETWEEN_PARENTHESIS, remaining)
    while match:
        extracted.append(match.group(1))
        start, end = match.span()
        remaining = remaining[:start] + remaining[end:]
        match = re.search(_PATTERN_BETWEEN_PARENTHESIS, remaining)
    return remaining, extracted



def pre_process(text: str) -> str:
    text = text.replace('‘', "'")
    text = text.replace('’', "'")
    text = text.replace('ʼ', "'")
    text = text.replace(',', ',')
    text = text.replace('˜', '') # Found instead of non-breakable spaces when parsing Ya! pdfs
    text = text.replace('Š', '') # Found instead of long dashes when parsing Ya! pdfs
    text = text.replace('ñ', 'ñ') # A sneaky n-tilde (found in Ya! webpages)
    text = text.replace('ň', 'ñ')
    text = text.replace('ù', 'ù') # Another sneaky one (found in Ya! webpages)
    return text



def fix_clitic(text: str) -> str:
    """ Do not use ! """
    
    text = text.replace("d' ", "d'")
    # text = text.replace("n' ", "n'")

    text = text.replace("n'eus ", "'n eus ")
    text = text.replace("n'int ", "n' int ")
    text = text.replace("n'eo ", "n' eo ")
    text = text.replace("n'hon ", "n' hon ")
    text = text.replace("n'ez ", "n' ez ")
    text = text.replace("n'em ", "n' em ")
    text = text.replace("n'am ", "n' am ")
    text = text.replace("n'en ", "n' en ")
    text = text.replace("n'o ", "n' o ")
    text = text.replace("n'on ", "n' on ")
    text = text.replace("n'he ", "n' he ")
    text = text.replace("n'edo ", "n' edo ")
    text = text.replace("n'emañ ", "n' emañ ")
    text = text.replace("n'anavezan ", "n' anavezan ")
    text = text.replace("n'ouzer ", "n' ouzer ")
    text = text.replace("n'ouzon ", "n' ouzon ")
    # n'oc'h
    # n'omp



def sentence_stats(sentence: str) -> dict:
    """ Get statistics about a string """
    
    letter = 0
    decimal = 0
    upper = 0
    punct = 0
    blank = 0
    other = 0
    
    for c in sentence:
        if c.lower() in LETTERS or c in "'-":
            letter += 1
            if c.isupper():
                upper += 1
        elif c.isdecimal():
            decimal += 1
        elif c.isspace():
            blank += 1
        elif c in PUNCTUATION:
            punct += 1
        else:
            other += 1
        
        sentence = filter_out(sentence, PUNCTUATION)
    
    return {"letter": letter, "decimal": decimal, "upper": upper, "punct": punct,
            "blank": blank, "other": other, "words": len(sentence.split())}


#! /usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Create subtitles (a `srt` file) from an audio file
and a text file, using a vosk model

Author:  Gweltaz Duval-Guennoc

Usage: ./linennan.py audio_file text_file

"""

import sys
import datetime
import argparse

import re
import srt
import jiwer
import static_ffmpeg
from tqdm import tqdm

from anaouder.asr.models import load_model, DEFAULT_MODEL
from anaouder.asr.recognizer import transcribe_file_timecoded
from anaouder.asr.post_processing import verbal_fillers
from anaouder.text import (
    pre_process, filter_out,
    sentence_stats,
    tokenize, detokenize, normalize_sentence, split_sentences,
    PUNCTUATION
)
from anaouder.utils import read_file_drop_comments
from anaouder.version import VERSION



autocorrect = False


def main_linennan() -> None:
    """ linennan cli entry point """

    parser = argparse.ArgumentParser(
        prog = "Linennan",
        description = "Create a timecoded file (`srt` or `split` file) from a text and an audio file")
    parser.add_argument("audio_file")
    parser.add_argument("text_file")
    parser.add_argument("-m", "--model", default=DEFAULT_MODEL,
        help="Vosk model to use for decoding", metavar='MODEL_PATH')
    parser.add_argument("-t", "--type", choices=["srt", "seg"],
        help="file output type")
    parser.add_argument("-r", "--reformat", action="store_true",
        help="reformat text file using punctuation, to put one sentence per line")
    parser.add_argument("--keep-fillers", action="store_true",
        help="Keep verbal fillers ('euh', 'beñ', 'alors', 'kwa'...)")
    parser.add_argument("-o", "--output", help="write to a file")
    parser.add_argument("-d", "--debug", action="store_true",
        help="display debug information")
    parser.add_argument("-v", "--version", action="version", version=f"%(prog)s v{VERSION}")
    args = parser.parse_args()
    
    # Use static_ffmpeg instead of ffmpeg
    static_ffmpeg.add_paths()

    load_model(args.model)

    lines = read_file_drop_comments(args.text_file)
    
    # Remove metadata
    lines = [ re.sub(r"{.+?}", '', l) for l in lines ]
    lines = [ l for l in lines if l ]

    if args.reformat:
        # Use punctuation to reformat text to one sentence per line
        data = ''.join(lines)
        parts = data.split('\n\n')
        lines = []
        for part in parts:
            lines.extend(split_sentences(part, end=''))
        
    # Remove special labels <C'HOARZH>, <UNK>...
    cleaned_lines = [ re.sub(r"<[A-Z\']+?>", '', line) for line in lines ]

    cleaned_lines = [ filter_out(pre_process(line), PUNCTUATION + "><*").strip() for line in lines ]
    
    # Look for numbers in text file, normalize in that case
    normalize = False
    for line in cleaned_lines:
        if sentence_stats(line)["decimal"] > 0:
            normalize = True
            break
    if normalize:
        cleaned_lines = [ normalize_sentence(line, autocorrect=autocorrect) for line in cleaned_lines ]
    
    cleaned_lines = [ " ".join(line.lower().replace('-', ' ').split()) for line in cleaned_lines ]
    # print(cleaned_lines, file=sys.stderr)

    hyp = transcribe_file_timecoded(args.audio_file)
    if not args.keep_fillers:
        print("Removing verbal fillers", file=sys.stderr)
        hyp = [ tok for tok in hyp if tok["word"] not in verbal_fillers ]
    print("Alignment...")


    # Try to locate the best match for each sentence in transcription

    sentence_matches = []
    for sentence in cleaned_lines:
        match = []
        for i in range(len(hyp)):
            # Try to find a local minima for the CER by adding one word at a time
            best_score = 999
            offset = 1
            while i+offset <= len(hyp):
                hyp_window = hyp[i: i+offset]
                hyp_sentence = ' '.join( [t["word"].lower().replace('-', ' ') for t in hyp_window] )
                score = jiwer.cer(
                    filter_out(sentence, ' '),
                    filter_out(hyp_sentence, ' ')
                    )
                if score <= best_score:
                    best_score = score
                    best_span = (i, i+offset)
                    best_hyp = hyp_sentence
                else:
                    break
                offset += 1
            
            match.append( {"hyp": best_hyp, "score": best_score, "span": best_span} )
        
        match.sort(key=lambda x: x["score"])
        sentence_matches.append(match[0])


    # Infer the reliability of each location by checking its adjacent neighbours

    def get_prev_wi(idx):
        if idx <= 0:
            return 0
        return sentence_matches[idx-1]["span"][1]
    
    def get_next_wi(idx):
        if idx >= len(sentence_matches) - 1:
            return len(hyp)
        return sentence_matches[idx+1]["span"][0]
    
    reliability = []
    last_reliable_wi = 0
    for i, match in enumerate(sentence_matches):
        span = match["span"]
        is_pdn = get_prev_wi(i) == span[0] # is prev a direct neighbour ?
        is_ndn = get_next_wi(i) == span[1] # is next a direct neighbour ?
        if is_pdn or is_ndn:
            last_reliable_wi = span[1]
            r = 'o' # Semi-reliable
            if is_pdn and is_ndn:
                r = 'O' # Reliable
        elif span[0] > last_reliable_wi:
            r = '?' # Not sure...
        else:
            r = 'X' # Obviously wrong
        reliability.append(r)
        # print(f"{i} {span}\t{r}")
    

    # Now that we have a reliability vector, we can try to deduce
    # the state (right or wrong) of the locations with "unknown" reliability

    def get_prev_reliable_wi(idx):
        if idx <= 0:
            return 0
        for i in range(idx, 0, -1):
            if reliability[i-1] in ('O', 'o'):
                return sentence_matches[i-1]["span"][1]
        return 0

    def get_next_reliable_wi(idx):
        if idx >= len(sentence_matches) - 1:
            return len(hyp)
        for i in range(idx+1, len(reliability)):
            if reliability[i] in ('O', 'o'):
                return sentence_matches[i]["span"][0]
        return len(hyp)

    for i in range(len(sentence_matches)):
        # An "unknown" is validated if it's location is between the
        # previous reliable location and the next reliable location
        # It is deemed as "wrong" otherwise
        if reliability[i] == '?':
            span = sentence_matches[i]["span"]
            p_rel_wi = get_prev_reliable_wi(i)
            n_rel_wi = get_next_reliable_wi(i)
            if p_rel_wi < span[0] and span[1] < n_rel_wi:
                reliability[i] = 'o'
            else:
                reliability[i] = 'X'


    # Try to repair remaining wrong locations
    for i in range(1, len(sentence_matches)-1):
        if reliability[i] == 'X':
            if reliability[i-1] in ('O', 'o') and reliability[i+1] in ('O', 'o'):
                span = (get_prev_wi(i), get_next_wi(i))
                hyp_window = hyp[span[0]: span[1]]
                hyp_sentence = ' '.join( [t["word"].lower().replace('-', ' ') for t in hyp_window] )
                score = jiwer.cer(
                    filter_out(cleaned_lines[i], ' '),
                    filter_out(hyp_sentence, ' ')
                    )
                sentence_matches[i] = {"hyp": hyp_sentence, "score": score, "span": span}
                reliability[i] = '.'
    
    # Stupid segmentation of remaining "unknown" groups
    unk_clumps = []
    in_clump = False
    clump_start = 0
    for i in range(len(reliability)):
        if reliability[i] == 'X':
            if not in_clump:
                clump_start = i
                in_clump = True
        elif in_clump:
            unk_clumps.append( (clump_start, i) )
            in_clump = False
    
    for clump in unk_clumps:
        size = clump[1] - clump[0]
        span = ( get_prev_wi(clump[0]), get_next_wi(clump[1]-1) )
        n_words = span[1] - span[0]
        step = n_words / size
        print(clump, span)
        prev = span[0]
        for i in range(size):
            new_span = (prev, prev + round(step))
            print("  ", new_span)
            hyp_window = hyp[new_span[0]: new_span[1]]
            hyp_sentence = ' '.join( [t["word"].lower().replace('-', ' ') for t in hyp_window] )
            score = jiwer.cer(
                filter_out(cleaned_lines[clump[0]+i], ' '),
                filter_out(hyp_sentence, ' ')
                )
            sentence_matches[clump[0]+i] = {"hyp": hyp_sentence, "score": score, "span": new_span}
            reliability[clump[0]+i] = '/'
            prev += round(step)
        
    

    total_cer = 0
    for i, match in enumerate(sentence_matches):
        total_cer += match["score"]
        if args.debug:
            span = match["span"]
            r = reliability[i]
            print(f"{i} {span}\t{r}", file=sys.stderr)
    print(f"Mean CER: {total_cer / len(sentence_matches):0.3}", file=sys.stderr)

    n = 0
    for i, l in enumerate(cleaned_lines):
        if reliability[i] in ('X', '/'):
            n += 1
            hyp_sentence = sentence_matches[i]["hyp"]
            if args.debug:
                print(f"{i} {sentence_matches[i]['span']}\t{l}\t{hyp_sentence}\t", file=sys.stderr)
    print(f"{n} ill-aligned segment{'s' if n>1 else ''}", file=sys.stderr)



    # Resolve file output type

    if args.output and not args.type:
        # No type explicitely defined, use output file extension
        split_ext = args.output.rsplit('.', maxsplit=1)
        if len(split_ext) == 2:
            ext = split_ext[1].lower()
            if ext in ("srt", "seg"):
                args.type = ext
            else:
                print("Unrecognized extension, using default type (`srt`)", file=sys.stderr)
                args.type = 'srt'
        else:
            # No file extension found, use default type
            args.type = 'srt'

    if not args.type:
        args.type = "srt"
    

    fout = open(args.output, 'w') if args.output else sys.stdout

    if args.type == "srt":
        subs = []
        last = -1
        for i, line in enumerate(lines):
            if reliability[i] == 'X':
                continue

            span = sentence_matches[i]["span"]
            start = hyp[span[0]]["start"]
            if start <= last: # Avoid timecode overlap
                start += 0.05
            end = hyp[span[1]-1]["end"]
            last = end
            s = srt.Subtitle(index=len(subs),
                    content=line,
                    start=datetime.timedelta(seconds=start),
                    end=datetime.timedelta(seconds=end))
            subs.append(s)

        print(srt.compose(subs), file=fout)
    

    if args.type == "seg":
        # utts = []
        for i, line in enumerate(lines):
            span = sentence_matches[i]["span"]
            start = int(hyp[span[0]]["start"] * 1000)
            end = int(hyp[span[1]-1]["end"] * 1000)
            print(f"{start} {end}", file=fout)
    
    if args.output:
        fout.close()



if __name__ == "__main__":
    main_linennan()



from typing import List

import os
import platform
import urllib
import zipfile
from tqdm import tqdm

from vosk import Model, SetLogLevel


DEFAULT_MODEL = "vosk-model-br-0.9"

_MODELS = {
    "vosk-model-br-0.6": "https://github.com/gweltou/vosk-br/releases/download/v0.6/vosk-model-br-0.6.zip",
    "vosk-model-br-0.7": "https://github.com/gweltou/vosk-br/releases/download/v0.7/vosk-model-br-0.7.zip",
    "vosk-model-br-0.8": "https://github.com/gweltou/vosk-br/releases/download/v0.8/vosk-model-br-0.8.zip",
    "vosk-model-br-0.9": "https://github.com/gweltou/vosk-br/releases/download/v0.9/vosk-model-br-0.9.zip",
}

_MODELS_ALIASES = {
    "vosk-br-0.6": "vosk-model-br-0.6",
    "vosk6": "vosk-model-br-0.6",
    "vosk-br-0.7": "vosk-model-br-0.7",
    "vosk7": "vosk-model-br-0.7",
    "vosk-br-0.8": "vosk-model-br-0.8",
    "vosk8": "vosk-model-br-0.8",
    "vosk-br-0.9": "vosk-model-br-0.9",
    "vosk9": "vosk-model-br-0.9",
    "vosk": "vosk-model-br-0.9",
}

_loaded_model = None
_loaded_model_name = ""


def available_models() -> List[str]:
    """ Returns the names of available models
        Code modified from https://github.com/openai/whisper
    """
    return list(_MODELS.keys())


def load_model(
    model_name: str = None,
    download_root: str = None,
) -> Model:
    """Code modified from https://github.com/openai/whisper"""
    global _loaded_model_name
    global _loaded_model
    
    if model_name == None:
        if _loaded_model:
            return _loaded_model
        model_name = DEFAULT_MODEL
    
    if model_name == _loaded_model_name:
        return _loaded_model

    if model_name in _MODELS_ALIASES:
        model_name = _MODELS_ALIASES[model_name]

    if download_root is None:
        if platform.system() in ("Linux", "Darwin"):
            default = os.path.join(os.path.expanduser("~"), ".cache")
        elif platform.system() == "Windows":
            default = os.getenv("LOCALAPPDATA")
        else:
            raise OSError('Unsupported operating system')
        download_root = os.path.join(os.getenv("XDG_CACHE_HOME", default), "anaouder", "models")

    if model_name in _MODELS:
        model_path = _download(model_name, download_root)
    elif os.path.isdir(model_name):
        model_path = model_name
    else:
        raise RuntimeError(
            f"Model {model_name} not found; available models = {available_models()}"
        )

    print("Loading", os.path.basename(model_path))
    SetLogLevel(-1)
    _loaded_model = Model(model_path)
    _loaded_model_name = model_name

    return _loaded_model


def _download(model_name: str, root: str) -> str:
    """ Code modified from https://github.com/openai/whisper
        Get the requested model path on disk or download it if not present
    """
    os.makedirs(root, exist_ok=True)
    url = _MODELS[model_name]
    
    download_target = os.path.join(root, os.path.basename(url))
    model_path = os.path.join(root, model_name)

    if os.path.isdir(model_path):
        return model_path

    print("Downloading model from", url)

    with urllib.request.urlopen(url) as source, open(download_target, "wb") as output:
        with tqdm(
            total=int(source.info().get("Content-Length")),
            ncols=80,
            unit="iB",
            unit_scale=True,
            unit_divisor=1024,
        ) as loop:
            while True:
                buffer = source.read(8192)
                if not buffer:
                    break

                output.write(buffer)
                loop.update(len(buffer))
    
    with zipfile.ZipFile(download_target, 'r') as zip_ref:
        zip_ref.extractall(root)

    os.remove(download_target)

    return model_path



#! /usr/bin/env python3

import sys
import os.path
import datetime
from math import floor, ceil
import argparse
import subprocess

import static_ffmpeg
import json
import srt
from vosk import KaldiRecognizer
from pydub import AudioSegment

from anaouder.asr.models import load_model, DEFAULT_MODEL
from anaouder.asr.recognizer import transcribe_file_timecoded, transcribe_segment_timecoded
from anaouder.asr.post_processing import post_process_text, post_process_timecoded
from anaouder.audio import split_to_segments, convert_to_wav
from anaouder.utils import write_eaf
from anaouder.version import VERSION



def _split_vosk_tokens(tokens, min_silence=0.5):
    subsegments = []
    current_segment = [tokens[0]]
    for tok in tokens[1:]:
        if tok['start'] - current_segment[-1]['end'] > min_silence:
            # We shall split here
            subsegments.append(current_segment)
            current_segment = []
        current_segment.append(tok)
    subsegments.append(current_segment)
    return subsegments


def split_vosk_tokens(tokens, max_length=15):
    """ Split sequences of Vosk tokens on silences """
    token_segments = [tokens]
    silence_length = 1.0

    while True:
        n_long_segments = 0
        parsed = []
        for segment in token_segments:
            dur = segment[-1]['end'] - segment[0]['start']
            if dur > max_length:
                # Split this segment deeper
                sub = _split_vosk_tokens(segment, silence_length)
                parsed.extend(sub)
                n_long_segments += 1
            else:
                parsed.append(segment)
        token_segments = parsed
        silence_length -= 0.1
        if n_long_segments == 0 or silence_length < 0.3:
            break
    
    return token_segments


def main_adskrivan(*args, **kwargs) -> None:
    """ adskrivan cli entry point """

    desc = f"Decode an audio file in any format, with the help of ffmpeg"
    parser = argparse.ArgumentParser(description=desc, prog="adskrivan")
    parser.add_argument('filename')
    parser.add_argument("-m", "--model", default=DEFAULT_MODEL,
        help="Vosk model to use for decoding", metavar='MODEL_PATH')
    parser.add_argument("-n", "--normalize", action="store_true",
        help="Normalize numbers")
    parser.add_argument("--keep-fillers", action="store_true",
        help="Keep verbal fillers ('euh', 'beñ', 'alors', 'kwa'...)")
    parser.add_argument("-t", "--type", choices=["txt", "srt", "eaf", "seg"],
        help="file output type")
    parser.add_argument("-o", "--output", help="write to a file")
    parser.add_argument("--autosplit", action="store_true",
        help="Automatically split audio at silences (used with 'srt', 'eaf' or 'seg' type only)")
    parser.add_argument("--segment-min-length",
        help="Will try not to go under this length when segmenting audio files (seconds)",
        type=float, default=2)
    parser.add_argument("--segment-max-length",
        help="Will try not to go above this length when segmenting audio files (seconds)",
        type=float, default=8)
    parser.add_argument("--max_words_per_line", type=int, default=7,
        help="Number of words per line for subtitle files")
    parser.add_argument("--set-ffmpeg-path", type=str,
        help="Set ffmpeg path (will not use static_ffmpeg in that case)")
    parser.add_argument("-v", "--version", action="version", version=f"%(prog)s v{VERSION}")

    if args:
        args = parser.parse_args(args)
    else:
        args = parser.parse_args()
    
    if not os.path.exists(args.filename):
        print("Couldn't find file '{}'".format(args.filename))
        sys.exit(1)
        
    
    # Use static_ffmpeg instead of ffmpeg
    ffmpeg_path = "ffmpeg"
    if args.set_ffmpeg_path:
        ffmpeg_path = args.set_ffmpeg_path
    else:
        static_ffmpeg.add_paths()
    

    if args.type == "seg":
        if args.output:
            args.output = args.output.replace(' ', '_')
        else:
            args.output = args.filename.replace(' ', '_')
            ext = os.path.splitext(args.output)[1]
            args.output = args.output.replace(ext, ".seg")
    
    fout = open(args.output, 'w') if args.output else sys.stdout

    if not args.type:
        if args.output:
            # No explicit type was given to we'll try to infer it from output file extension
            ext = os.path.splitext(args.output)[1][1:]
            if ext.lower() in ("srt", "seg", "eaf"):
                args.type = ext.lower()
            else:
                args.type = "txt"	# Default type
        else:
            args.type = "txt"

    model = load_model(args.model)

    if args.type == "txt":
        # No need of timecodes

        if args.output:
            # Whole file decoding
            # Different segmentation algorithm than online decoding
            # Shows a progress bar
            print("Transcribing audio file...", flush=True)
            tokens = transcribe_file_timecoded(args.filename)
            token_segments = split_vosk_tokens(tokens, max_length=args.segment_max_length)
            token_segments = [
                post_process_timecoded(seg, args.normalize, args.keep_fillers)
                for seg in token_segments
            ]
            sentences = [ ' '.join([token['word'] for token in seg]) for seg in token_segments ]

            fout.write('\n'.join(sentences))
        else:
            # Online decoding
            # Print decoded sentences one-by-one
            rec = KaldiRecognizer(model, 16000)
            rec.SetWords(True)
        
            with subprocess.Popen([ffmpeg_path, "-loglevel", "quiet", "-i",
                                    args.filename,
                                    "-ar", "16000" , "-ac", "1", "-f", "s16le", "-"],
                                    stdout=subprocess.PIPE).stdout as stream:
                while True:
                    data = stream.read(4000)
                    if len(data) == 0:
                        break
                    if rec.AcceptWaveform(data):
                        sentence = json.loads(rec.Result())["text"]
                        sentence = post_process_text(sentence, normalize=args.normalize, keep_fillers=args.keep_fillers)
                        if sentence: print(sentence, file=fout)

                sentence = json.loads(rec.FinalResult())["text"]
                sentence = post_process_text(sentence, normalize=args.normalize, keep_fillers=args.keep_fillers)
                if sentence: print(sentence, file=fout)


    elif args.type in ("srt", "seg", "eaf"):
        # Transcribe with timecodes
        
        if args.autosplit:
            song = AudioSegment.from_file(args.filename)
            song = song.set_channels(1)
            if song.frame_rate != 16000:
                song = song.set_frame_rate(16000)
            if song.sample_width != 2:
                song = song.set_sample_width(2)
            
            # Audio need to be segmented first
            print("Segmenting audio file...", end=' ', flush=True)
            segments = split_to_segments(
                song,
                max_length=args.segment_max_length,
                min_length=args.segment_min_length)
            print(f"{len(segments)} segment{'s' if len(segments)>1 else ''} found")

            t_min, t_max = 0, segments[-1][1]
            sentences = []
            for seg in segments:
                tr = transcribe_segment_timecoded(song[max(t_min, seg[0]-200):min(t_max, seg[1]+200)])
                sentences.append(tr)
            
        else:
            print("Transcribing audio file...", flush=True)
            tokens = transcribe_file_timecoded(args.filename)
            sentences = split_vosk_tokens(tokens, max_length=args.segment_max_length)
            segments = [ [floor(sent[0]['start']*1000), ceil(sent[-1]['end']*1000)]
                               for sent in sentences ]

        # Remove empty sentences
        new_sentences = [s for s in sentences if s]
        new_segments = [segments[i] for i, s in enumerate(sentences) if s]
        segments = new_segments
        sentences = new_sentences

        # Apply text post-processing
        if args.type == "seg":
            args.keep_fillers = True
        sentences = [
            post_process_timecoded(sent, args.normalize, args.keep_fillers)
            for sent in sentences
        ]


        if args.type == 'seg':
            for start, end in segments:
                print(f"{start} {end}", file=fout)
            
            textfile_header = \
            "{source: }\n{source-audio: }\n{author: }\n{licence: }\n{tags: }\n\n\n\n"
            
            if args.output:
                basename = os.path.splitext(args.output)[0]
            else:
                basename = os.path.splitext(args.filename)[0]
            text_filename = os.path.extsep.join([basename, 'txt'])
            
            with open(text_filename, 'w') as fw:
                fw.write(textfile_header)  # Text file split_header
                for s in sentences:
                    sentence = ' '.join([token['word'] for token in s])
                    fw.write(sentence + '\n')
            
            wav_filename = os.path.extsep.join([basename, 'wav'])
            print(args.filename, wav_filename)
            convert_to_wav(args.filename, wav_filename)
        
        
        elif args.type == 'srt':
            words_per_line = args.max_words_per_line
            if words_per_line == 0: words_per_line = 999
            subs = []
            for i, sentence in enumerate(sentences):
                for j in range(0, len(sentence), words_per_line):
                    line = sentence[j : j + words_per_line]
                    text = ' '.join([ w["word"] for w in line ])

                    if args.autosplit:
                        offset = segments[i][0]/1000
                        s = srt.Subtitle(index=len(subs),
                                content=text,
                                start=datetime.timedelta(seconds=line[0]["start"]+offset),
                                end=datetime.timedelta(seconds=line[-1]["end"]+offset))
                    else:
                        s = srt.Subtitle(index=len(subs),
                                content=text,
                                start=datetime.timedelta(seconds=line[0]["start"]),
                                end=datetime.timedelta(seconds=line[-1]["end"]))
                    subs.append(s)
            
            # Write to a srt file with the same name as input file by default
            if not args.output:
                srt_path = os.path.splitext(args.filename)[0] + ".srt"
                fout = open(srt_path, 'w')

            print(srt.compose(subs), file=fout)
        
        
        elif args.type == 'eaf':
            text_sentences = []
            for sentence in sentences:
                sentence = ' '.join([ t['word'] for t in sentence ])
                text_sentences.append(sentence)
            ext = os.path.splitext(args.output)[1][1:].lower()
            if ext not in ('mp3', 'wav'):
                data = write_eaf(segments, text_sentences, args.filename, type="mp3")
            else:
                data = write_eaf(segments, text_sentences, args.filename)

            print(data, file=fout)
            if args.output:
                print("EAF file written to", os.path.abspath(args.output))

  
    if args.output:
        fout.close()
    

if __name__ == "__main__":
    main_adskrivan()


#! /usr/bin/env python3
# -*- coding: utf-8 -*-

import sys
import argparse
from anaouder.text import normalize_sentence
from anaouder.version import VERSION



def main_normalizan(*args, **kwargs):
    """ normalizan cli entry point """
    
    parser = argparse.ArgumentParser(
        prog = "normalizan",
        description = "Normalize a text file")
    parser.add_argument("-i", "--input", help="Input file")
    parser.add_argument("-o", "--output", help="write to a file")
    parser.add_argument("-r", "--reformat", action="store_true",
        help="reformat text file using punctuation, to put one sentence per line")
    parser.add_argument("-v", "--version", action="version", version=f"%(prog)s v{VERSION}")
    args = parser.parse_args()
    
    input = open(args.input, 'r') if args.input else sys.stdin
    output = open(args.output, 'w') if args.output else sys.stdout

    while True:
        data = input.readline()
        if data:
            print(normalize_sentence(data), file=output)
        else:
            break
    
    if args.input:
        input.close()
    if args.output:
        output.close()


if __name__ == "__main__":
    main_normalizan()



from typing import List
import re
from ..dicts import proper_nouns, nouns_f, nouns_m


LETTERS = "aâàbcçdeêéèfghiïjklmnñoôpqrstuüùûvwxyz"
PUNCTUATION = '.?!,‚;:«»“”"()[]/…–—'
OPENING_QUOTES = "«“"
CLOSING_QUOTES = "»”"
# CLOSING_PUNCT = {'»': '«', '”': '“', ')': '('}
# OPENING_PUNCT = CLOSING_PUNCT.values()


# Acronyms
PATTERN_DOTTED_ACRONYM = re.compile(r"([A-Z]\.)+([A-Z])?")



SI_UNITS = {
    'g'     : ["gramm"],
    'kg'    : ["kilo", "kilogramm"],
    't'     : ["tonenn"],
    'l'     : ["litr", "litrad"],
    'cl'    : ["santilitr", "santilitrad"],
    'ml'    : ["mililitr", "mililitrad"],
    'cm'    : ["santimetr", "kantimetr"],
    'cm2'   : ["santimetr karrez", "kantimetr karrez"],
    'cm²'   : ["santimetr karrez", "kantimetr karrez"],
    'cm3'   : ["santimetr diñs", "kantimetr diñs"],
    'm'     : ["metr", "metrad"],
    'm2'    : ["metr karrez", "metrad karrez"],
    'm²'    : ["metr karrez", "metrad karrez"],
    'm3'    : ["metr diñs", "metrad diñs"],
    'm³'    : ["metr diñs", "metrad diñs"],
    'km'    : ["kilometr", "kilometrad"],
    "c'hm"  : ["c'hilometr", "c'hilometrad"],
    'km2'   : ["kilometr karrez", "kilometrad karrez"],
    'km²'   : ["kilometr karrez", "kilometrad karrez"],
    'km³'   : ["kilometr diñs", "kilometrad diñs"],
    'mn'    : ["munutenn"],
    '€'     : ['euro'],
    '$'     : ['dollar', 'dollar amerikan'],
    'M€'    : ['milion euro'],
    '%'     : ["dre gant"],
    }

# A percentage or a number followed by a unit

re_unit_number = re.compile(r"(\d+)([\w%€$']+)", re.IGNORECASE)
match_unit_number = lambda s: re_unit_number.fullmatch(s)
def is_unit_number(s):
    match = match_unit_number(s)
    if not match: return False
    unit = match.group(2)
    return unit in SI_UNITS


# Time (hours and minutes)

re_time = re.compile(r"(\d+)(?:e|h|eur)(\d+)?")
match_time = lambda s: re_time.fullmatch(s)
def is_time(s):
    match = match_time(s)
    if not match: return False
    _, m = match.groups(default='00')
    return int(m) < 60


# Nouns and proper nouns

def is_proper_noun(word: str) -> bool:
    if '-' in word:
        for sub in word.split('-'):
            if sub not in proper_nouns:
                return False
        return True
    else:
        return word in proper_nouns


def is_noun_f(word: str) -> bool:
    if len(word) < 2:
        return False
    
    word = word.lower()
    if word in nouns_f:
        return True
    for candidate in reverse_mutation(word):
        if candidate in nouns_f:
            return True
    return False


def is_noun_m(word: str) -> bool:
    if len(word) < 2:
        return False

    word = word.lower()
    if word in nouns_m:
        return True
    for candidate in reverse_mutation(word):
        if candidate in nouns_m:
            return True
    return False


def is_noun(word: str) -> bool:
    return is_noun_m(word) or is_noun_f(word)



def reverse_mutation(word: str) -> List[str]:
    """ returns the list of possible reverse-mutated candidates
        from a mutated (or not) word.
        Note that many candidates won't have meaning
    """
    first_letter = word[0].lower()
    is_cap = word[0].isupper()
    candidates = []

    if word.lower().startswith("c'h"):
        candidates.append('k' + word[3:])
        candidates.append('g' + word[3:])
    elif first_letter == 'w':
        candidates.append('g' + word[:])
    elif first_letter == 'z':
        candidates.append('t' + word[1:])
        candidates.append('d' + word[1:])
    elif first_letter == 'f':
        candidates.append('p' + word[1:])
    elif first_letter == 'k':
        candidates.append('g' + word[1:])
    elif first_letter == 't':
        candidates.append('d' + word[1:])
    elif first_letter == 'p':
        candidates.append('b' + word[1:])
    elif first_letter == 'g':
        candidates.append('k' + word[1:])
    elif first_letter == 'd':
        candidates.append('t' + word[1:])
    elif first_letter == 'b':
        candidates.append('p' + word[1:])
    elif first_letter == 'v':
        candidates.append('b' + word[1:])
        candidates.append('m' + word[1:])
    
    if is_cap:
        candidates = [ w[0].upper() + w[1:] for w in candidates]
    return candidates



from os import path
from setuptools import setup, find_packages
from anaouder.version import VERSION


NAME = "anaouder"
DESCRIPTION = "Breton language speech-to-text tools"
URL = "https://github.com/gweltou/vosk-br"
AUTHOR = "Gweltaz Duval-Guennoc"
EMAIL = "gweltou@hotmail.com"
REQUIRES_PYTHON = ">=3.6.0"


# The directory containing this file
HERE = path.dirname(__file__)

with open(path.join(HERE, "requirements.txt")) as fd:
    REQUIREMENTS = [line.strip() for line in fd.readlines() if line.strip()]


setup(
    name=NAME,
    url=URL,
    version=VERSION,
    author=AUTHOR,
    license="MIT",
    author_email=EMAIL,
    description=DESCRIPTION,
    long_description=open("README-fr.md", encoding="utf-8").read(),
    long_description_content_type="text/markdown",
    python_requires=REQUIRES_PYTHON,
    install_requires=REQUIREMENTS,
    classifiers=[
		"Intended Audience :: Developers",
		"License :: OSI Approved :: MIT License",
		"Programming Language :: Python",
		"Programming Language :: Python :: 3",
		"Programming Language :: Python :: 3.6",
		"Programming Language :: Python :: 3.7",
		"Programming Language :: Python :: 3.8",
		"Programming Language :: Python :: 3.9",
		"Operating System :: OS Independent"
	],
    packages=find_packages(),
    # package_dir={"anaouder": "anaouder", "ostilhou": "anaouder/ostilhou"},
    package_data={
        "anaouder": [
            "asr/inorm_units.tsv",
            "asr/postproc_sub.tsv",
            "dicts/acronyms.tsv",
            "dicts/corrected_tokens.tsv",
            "dicts/noun_f.tsv",
            "dicts/noun_m.tsv",
            "dicts/proper_nouns_phon.tsv",
            "dicts/standard_tokens.tsv",
        ]
    },
    data_files=[('', ["README.md", "README-fr.md", "CHANGELOG.md"])],
    entry_points={
        "console_scripts": [
            "adskrivan = anaouder:main_adskrivan",
            "linennan = anaouder:main_linennan",
            "istitlan = anaouder:main_istitlan",
            "mikro = anaouder:main_mikro",
            "normalizan = anaouder:main_normalizan",
        ],
    },
)



#!/usr/bin/env python3

import sys
from anaouder import main_adskrivan


def main_istitlan():
	args = sys.argv[1:] + ["-t", "srt"]
	return main_adskrivan(*args)


if __name__ == "__main__":
	main_istitlan()